/*************************************************************************/
/* mie.c                                                                 */
/*                                                                       */
/* mie.a - Mie calculations                                        @ti@  */
/*                                                                       */
/* Author:                                                               */
/*   Bernhard Mayer and Arve Kylling                               @au@  */
/*                                                                       */
/* ----------------------------------------------------------------------*/
/* Copyright (C) 2000 Bernhard Mayer and Arve Kylling                    */
/*                                                                       */
/* This program is free software; you can redistribute it and/or modify  */
/* it under the terms of the GNU General Public License as published by  */
/* the Free Software Foundation; either version 1, or (at your option)   */
/* any later version.                                                    */
/*                                                                       */
/* This program is distributed in the hope that it will be useful,       */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/* MERCHANTABILITY of FITNESS FOR A PARTICULAR PURPOSE. See the          */
/* GNU General Public License for more details.                          */
/*                                                                       */
/* To obtain a copy of the GNU General Public License write to the       */
/* Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,    */
/* USA.                                                                  */
/*-----------------------------------------------------------------------*/
/*************************************************************************/

/* @44c@ */
/* @c44@ */

/********************************************************************************/
/* In order to use the functions provided by the mie library,          @44_10c@ */
/* #include <miecalc.h> in your source code and link with libRadtran_c.a.       */
/*                                                                              */
/* @strong{Example:}                                                            */
/* Example for a source file:                                                   */
/* @example                                                                     */
/*                                                                              */
/*   ...                                                                        */
/*   #include "../src_c/miecalc.h"                                              */
/*   ...                                                                        */
/*                                                                              */
/* @end example                                                                 */
/*                                                                              */
/* Linking of the executable, using the GNU compiler gcc:                       */
/* @example                                                                     */
/*                                                                              */
/*   gcc -o test test.c -lRadtran_c                                             */
/*                                                                              */
/* @end example                                                        @c44_10@ */
/********************************************************************************/

/****************************************************************************/
/* The Mie library provides functions for Mie calculations,       @44_20c@  */
/* interfacing the MIEV0 and BHMIE codes by Warren Wiscombe                 */
/* (ftp://climate.gsfc.nasa.gov/pub/wiscombe) and Bohren and Huffman        */
/* (ftp://astro.princeton.edu/draine/scat/bhmie/). Functions for            */
/* evaluating the phase function and the integrated phase function are      */
/* also provided.                                                 @c44_20@  */
/****************************************************************************/


#include "../src_f/f77-uscore.h"
#include "../src_c/fortran_and_c.h"
#include "./numeric.h"
#include "./miecalc.h"

#ifndef PI
#define PI 3.14159265358979323846264338327
#endif


/* density of water in g / m3 */
#define RHOH2O 0.998E+6  




/***********************************************************************************/
/* Function: mie_calc                                                     @44_30i@ */
/* Description:                                                                    */
/*  Mie calculations, using the MIEV0 and BHMIE codes by Warren Wiscombe           */
/*  (ftp://climate.gsfc.nasa.gov/pub/wiscombe) and Bohren and Huffman              */
/*  (ftp://astro.princeton.edu/draine/scat/bhmie/).                                */
/*                                                                                 */
/* Parameters:                                                                     */
/*  mie_inp_struct input:    mie input structure (see src_c/miecalc.h)             */
/*  mie_out_struct *output:  mie output structure (see src_c/miecalc.h)            */
/*  int program:             MIEV0 or BHMIE                                        */
/*  int medium:              WATER, ICE, or USER; if USER, the refractive index    */
/*                           is read from crefin                                   */
/*  complex crefin:          Complex refractive index (both numbers positive)      */
/*  float wavelength:        Wavelength [micron]                                   */
/*  float radius:            Droplet radius [micron]                               */
/*  float temperature:       Temperature [K]                                       */
/*  complex *ref:            Complex index of refraction                           */
/*                                                                                 */
/* Return value:                                                                   */
/*  0  if o.k., <0 if error                                                        */
/*                                                                                 */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/*  Syntax and parameters of this function are subject to change.                  */
/* Author:                                                                         */
/*                                                                        @i44_30@ */
/***********************************************************************************/

int mie_calc (mie_inp_struct input, mie_out_struct *output, 
	      int program, int medium, complex crefin, 
	      float wavelength, float radius, 
	      float temperature, complex *ref)
{

  complex F77_FCN (refice) (float *wl, float *temp); 
  complex F77_FCN (refwat) (float *wl, float *temp); 
  void F77_FCN (bhmie) (float *xx, complex *crefin, int *numang,
			complex *s1, complex *s2,  
			float *qext, float *qsca, float *qback, float *gsca);
  void F77_FCN (miev0) ( float *xx, complex *crefin, int *perfct, float *mimcut, 
			 int *anyang, int *numang, float *xmu, int *nmom, 
			 int *ipolzn, int *momdim, int *prnt, float *qext, 
			 float *qsca, float *gqsc, float *pmom, complex *sforw, 
			 complex *sback, complex *s1, complex *s2, complex *tforw, 
			 complex *tback, float *spike);
  
  float *tmp_pmom = (float *) calloc ((input.nmom+1)*4, sizeof(float));

  switch (medium) {
  case ICE:
    *ref = F77_FCN (refice) (&wavelength, &temperature);
    output->crefin = *ref;
    break;
    
  case WATER:
    *ref = F77_FCN (refwat) (&wavelength, &temperature);
    output->crefin = *ref;
    break;

  case USER:
    ref->re = crefin.re;
    ref->im = crefin.im;
    output->crefin.re = crefin.re;
    output->crefin.im = crefin.im;
    break;
    
  default:
    fprintf (stderr, "error: unknown medium\n");
    return -1;
  }
  
  
  output->xx     = 2.0*PI * radius / wavelength;

  switch (program) {
  case BHMIE:
    F77_FCN (bhmie) ( &output->xx, &output->crefin, &input.numang,
		      input.s1, input.s2,  
		      &output->qext, &output->qsca, &output->qback, 
		      &output->gsca );
    break;
  case MIEV0:
    F77_FCN (miev0) ( &output->xx, &output->crefin, &input.perfct, 
		      &input.mimcut, &input.anyang, &input.numang,  
		      input.xmu, &input.nmom, &input.ipolzn,  
		      &input.momdim, input.prnt,  
		      &output->qext, &output->qsca, &output->gqsc,  
		      tmp_pmom, &output->sforw, &output->sback,  
		      input.s1, input.s2,  
		      output->tforw, output->tback, &output->spike ); 
    
    output->gsca = output->gqsc / output->qsca;

    fortran2c_2D_float_ary_noalloc (4, input.nmom+1, tmp_pmom, output->pmom);
    
    break;
  default:
    fprintf (stderr, "error: unknown Mie program\n");
    return -1;
  }

  free (tmp_pmom);

  return 0;
}





/***********************************************************************************/
/* Function: mie_calc_sizedist                                            @44_30i@ */
/* Description:                                                                    */
/*  Mie calculations, using the MIEV0 and BHMIE codes by Warren Wiscombe           */
/*  (ftp://climate.gsfc.nasa.gov/pub/wiscombe) and Bohren and Huffman              */
/*  (ftp://astro.princeton.edu/draine/scat/bhmie/).                                */
/*                                                                                 */
/* Parameters:                                                                     */
/*  mie_inp_struct input:    mie input structure (see src_c/miecalc.h)             */
/*  mie_out_struct *output:  mie output structure (see src_c/miecalc.h)            */
/*  int program:             MIEV0 or BHMIE                                        */
/*  int medium:              WATER, ICE, or USER; if USER, the refractive index    */
/*                           is read from crefin                                   */
/*  complex crefin:          Complex refractive index (both numbers positive)      */
/*  float wavelength:        wavelength [micron]                                   */
/*  float temperature:       temperature                                           */
/*  double *x_size:          size distribution, radius [um]                        */
/*  double *y_size:          size distribution, n(r)                               */
/*  int n_size:              size distribution, number of radii                    */
/*  double *beta:            extinction coefficient [km-1] per unit                */
/*                           liquid water content (returned)                       */
/*  double *omega:           Single scattering albedo (returned)                   */
/*  double *g:               Asymmetry parameter (returned)                        */
/*                                                                                 */
/* Return value:                                                                   */
/*  0  if o.k., <0 if error                                                        */
/*                                                                                 */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/*  Syntax and parameters of this function are subject to change.                  */
/* Author:                                                                         */
/*                                                                        @i44_30@ */
/***********************************************************************************/

int mie_calc_sizedist (mie_inp_struct input, mie_out_struct *output, 
		       int program, int medium, complex crefin, 
		       float wavelength, float temperature, 
		       double *x_size, double *y_size, int n_size,
		       double *beta, double *omega, double *g)
{
  int status=0, is=0, ip=0, im=0;
  complex ref;

  double xsquared=0, xcubed=0, norm=0;
  
  double ext=0, sca=0, asy=0, vol=0;

  /* allocate memory for fields */
  double *extinction = calloc (n_size, sizeof (double));
  double *scattering = calloc (n_size, sizeof (double));
  double *asymmetry  = calloc (n_size, sizeof (double));
  double *volume     = calloc (n_size, sizeof (double));

  double ***pmom=NULL;

  if (input.nmom>0) {
    pmom     = calloc (4, sizeof(double **));
    
    for (ip=0; ip<4; ip++) {
      pmom    [ip] = calloc (input.nmom+1, sizeof(double *));
      
      for (im=0; im<=input.nmom; im++)
	pmom[ip][im] = calloc (n_size, sizeof(double));
    }
  }

  
  for (is=0; is<n_size; is++) {

    status = mie_calc (input, output, program, medium, crefin, 
		       wavelength, x_size[is], temperature, &ref);
    if (status!=0) {
      fprintf (stderr, "error %d returned by mie_calc()\n", status);
      return status;
    }
    
    xsquared = x_size[is] * x_size[is] * 1e-12;
    xcubed   = xsquared   * x_size[is] * 1e-6;
    
    extinction[is] = output->qext * xsquared * y_size[is];
    scattering[is] = output->qsca * xsquared * y_size[is];
    asymmetry [is] = output->gsca * scattering[is]; 
    volume    [is] = xcubed * y_size[is];

	
    /* quote from MIEV.doc, concerning the normalization of pmom[]:   */
    /*	                                                              */
    /*	The normalized moments are                                    */
    /*      4 / ( XX**2 * QSCA ) * PMOM,  but it is PMOM itself, not  */
    /*      these normalized moments, which should be integrated over */
    /*      a size distribution.                                      */
    
    if (input.nmom>0)
      for (ip=0; ip<4; ip++)
	for (im=0; im<=input.nmom; im++)
	      pmom[ip][im][is] = output->pmom[ip][im] * y_size[is];
  }
  
  norm = integrate (x_size, y_size,     n_size);
  
  ext  = integrate (x_size, extinction, n_size) / norm;
  sca  = integrate (x_size, scattering, n_size) / norm;
  asy  = integrate (x_size, asymmetry,  n_size) / norm;
  vol  = integrate (x_size, volume,     n_size) / norm;
  
  if (input.nmom>0)
    for (ip=0; ip<4; ip++)
      for (im=0; im<=input.nmom; im++)
	output->pmom[ip][im] = integrate (x_size, pmom[ip][im], n_size) / norm;
  

  *beta  = 3.0*ext/(4.0*RHOH2O*vol)*1000.0;
  *omega = sca/ext;
  *g     = asy/sca;



  /* free memory */
  free (extinction);
  free (scattering);
  free (asymmetry);
  free (volume);
  
  if (input.nmom>0) {
    for (ip=0; ip<4; ip++) {
      for (im=0; im<=input.nmom; im++)
	free(pmom[ip][im]);
      
      free(pmom[ip]);
    }
    
    free(pmom);
  }


  return 0;
}




/************************************************************************************/
/* Function: phase_function                                               @44_30i@  */
/* Description:                                                                     */
/*  Calculate the phase function from its moments and output to a file "phase.dat". */
/*  The phase function                                                              */
/*    @iftex                                                                        */
/*    @tex                                                                          */
/*     $p(\mu)$ is @*                                                               */
/*      $$ p (\mu) = \sum_{m=0}^{\infty} (2m+1) \cdot k_m \cdot P_m (\mu) $$        */
/*      where $k_m$ is the m'th moment and $P_m (\mu)$ is the m'th                  */
/*      Legendre polynomial.                                                        */
/*     @end tex                                                                     */
/*    @end iftex                                                                    */
/*    @ifinfo                                                                       */
/*     p(mu) is @*                                                                  */
/*     p (mu) = sum (m=0 to infinity) (2m+1) * k(m) * Pm (mu) @*                    */
/*     where k(m) is the m'th moment and Pm (mu) is the m'th Legendre polynomial.   */
/*    @end ifinfo                                                                   */
/*                                                                                  */
/* Parameters:                                                                      */
/*  float *moment: moments of the phase function, f[0, ..., L-1]                    */
/*  int L:         number of moments                                                */
/*                                                                                  */
/* Return value:                                                                    */
/*  0  if o.k., <0 if error                                                         */
/*                                                                                  */
/* Example:                                                                         */
/* Files:                                                                           */
/* Known bugs:                                                                      */
/*  Syntax and parameters of this function are subject to change.                   */
/* Author:                                                                          */
/*                                                                        @i44_30@  */
/************************************************************************************/

int phase_function (float *moment, int L)
{
  int l=0, it=0;
  double p0=0, p1=0, p2=0;
  double mu=0, dmu=0;
  double sum=0, sum2=0, factor=0;
  int Nt=0;
  FILE *outfile=fopen("phase.dat", "w");


  Nt  = 100001;
  dmu = 2.0 / (double) (Nt-1); 
  
  factor = 1.0 / 4.0 / 3.141692653590 / moment[0];

  for (it=0; it<Nt; it++) {

    mu = -1.0 + (double) it * dmu;
    
    p0=1;
    p1=mu;
    

    sum  = moment[0]*p0 + 3.0*moment[1]*p1;

    for (l=2; l<=L; l++) {
      p2 = (double) (2*l-1) / (double) l * mu * p1 - 
	(double) (l-1) / (double) l * p0;
      
      sum += (double) (2*l+1) * moment[l] * p2;
      p0=p1;
      p1=p2;
    }

    fprintf (outfile, "%g  %g  %g\n", mu, sum*factor, sum2*factor);
  }

  fclose (outfile);
  return 0;
}




/***********************************************************************************/
/* Function: cumulative_probability                                       @44_30i@ */
/* Description:                                                                    */
/*  Calculate the cumulative probability distribution from the moments of the      */
/*  phase function and output to file "cumulative.dat". The algorithm is described */
/*  by B.R. Barkstrom, "An efficient algorithm for choosing scattering directions  */
/*   in Monte Carlo work with arbitrary phase functions",                          */
/*   J.Q.S.R.T., 53, 23-38, 1995.                                                  */ 
/*  The phase function                                                             */
/*    @iftex                                                                       */
/*    @tex                                                                         */
/*     $p(\mu)$ is @*                                                              */
/*      $$ p (\mu) = \sum_{m=0}^{\infty} (2m+1) \cdot k_m \cdot P_m (\mu) $$       */
/*      where $k_m$ is the m'th moment and $P_m (\mu)$ is the m'th                 */
/*      Legendre polynomial.                                                       */
/*     @end tex                                                                    */
/*    @end iftex                                                                   */
/*    @ifinfo                                                                      */
/*     p(mu) is @*                                                                 */
/*     p (mu) = sum (m=0 to infinity) (2m+1) * k(m) * Pm (mu) @*                   */
/*     where k(m) is the m'th moment and Pm (mu) is the m'th Legendre polynomial.  */
/*    @end ifinfo                                                                  */
/*                                                                                 */
/* Parameters:                                                                     */
/*  float *moment: moments of the phase function, f[0, ..., L-1]                   */
/*  int L:         number of moments                                               */
/*                                                                                 */
/* Return value:                                                                   */
/*  0  if o.k., <0 if error                                                        */
/*                                                                                 */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/*  Syntax and parameters of this function are subject to change.                  */
/* Author:                                                                         */
/*                                                                        @i44_30@ */
/***********************************************************************************/

int cumulative_probability (float *moment, int L)
{
  int l=0, it=0;

  double *f=NULL;

  int Nt=0;

  FILE *outfile=fopen("cumulative.dat", "w");

  double dmu=0, factor=0;

  double *mu=NULL, *F=NULL;

  double *a=NULL, *b=NULL, *q=NULL, *c=NULL;


  f = calloc (L+1, sizeof(double));

  /* multiply by (2l+1) to convert from 'moments of the phase function' */
  /* to Legendre coefficients;                                          */
  for (l=0; l<=L; l++)
    f[l] = moment[l] * (double) (2*l+1);

  /* normalize to 1/4pi */
  factor = 1.0 / 4.0 / PI / f[0];

  for (l=0; l<=L; l++)
    f[l] *= factor;


  /* allocate memory for coefficients */
  a = calloc (L+2, sizeof(double));
  b = calloc (L+2, sizeof(double));
  q = calloc (L+2, sizeof(double));

  c = calloc (L+3, sizeof(double));



  /* calculate coefficients */
  for (l=1; l<L+1; l++) {
    a[l] = (double) (2*l+1) / (double) (l+2);
    b[l] = (double) (-l+1)  / (double) (l+2);
    q[l] = 2.0 * PI * f[l];
  }

  b[l] = (double) (-l+1) / (double) (l+2);


  Nt  = 100001;
  dmu = 2.0 / (double) (Nt-1); 
  
  /* allocate memory for the cumulative probability table */
  F  = calloc (Nt, sizeof(double));
  mu = calloc (Nt, sizeof(double));
  

  for (it=0; it<Nt; it++) {
    
    mu[it] = -1.0 + (double) it * dmu;
    
    for (l=L; l>0; l--)
      c[l] = mu[it] * a[l]*c[l+1] + b[l+1]*c[l+2] + q[l];
    
    F[it] = (1.0 + mu[it]) * (1.0 - c[1] * (1.0 - mu[it])) / 2.0;

    fprintf (outfile, "%9.6f %.10e\n", mu[it], F[it]); 
  }

  fclose(outfile);

  /* free memory */
  free (F); free (mu);

  return 0;
}









