/************************************************************************/
/*                                                                      */
/* Copyright (C) 1997 Arve Kylling                                      */
/*                                                                      */
/* This program is free software; you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation; either version 1, or (at your option)  */
/* any later version.                                                   */
/*                                                                      */
/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY of FITNESS FOR A PARTICULAR PURPOSE. See the         */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* To obtain a copy of the GNU General Public License write to the      */
/* Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,   */
/* USA.                                                                 */
/*                                                                      */
/************************************************************************/

#include <stdio.h>							    
#include <stdlib.h>
#include <math.h>
#include "uvspec.h"
#include "../src_c/ascii.h"
#include "../src_c/miecalc.h"
#include "../src_c/fortran_and_c.h"
#include "../src_f/f77-uscore.h"



/************************************/
/* prototypes of internal functions */
/************************************/

static int read_aerosol_profile (char *filename, int nlyr, int nlambda_t, 
				 float *zd, float **dtau);
static int read_aerosol_moments (char *filename,
				 float **moment, int *nmom);
static int scale_aer_angstrom (float **dtau, float *lambda_t, int nlambda_t, int nlyr,
			       float alpha, float beta);


/**************************************************************/
/* Setup atmospheric aerosol.                                 */
/**************************************************************/

int setup_aerosol (input_struct input, output_struct *output)
{
  int status=0, ip=0, iv=0, lc=0, k=0, kmax=0;
  int maxawvn = MAXAWVN;

  int linear=0;
  int refrac=0, sizedist=0;

  float *tmp_aerext=NULL, *tmp_aerabs=NULL, *tmp_aersym=NULL, *tmp_aerwvn=NULL;  
  float *tmp_aer_dtau=NULL, *tmp_aer_gg=NULL, *tmp_aer_ssa=NULL;

  float *lref=NULL, *re=NULL, *im=NULL;
  int nref=0;

  double temperature = 300.0;  /* The temperature is only used for wavelengths > 167 um */
  double beta=0, beta0=0, ext=0, omega=0, g=0;
  
  complex crefin;

  mie_inp_struct *mie_inp=NULL;
  mie_out_struct *mie_out=NULL;


  

  int nmom=1000;   /* ??? should tie this to nstr to save computational time ??? */


  void F77_FCN (aerprof) (int* nlyr, float* zd, float* aer_pro, 
			  int* seasn, int* vulcan, float* visib );
  void F77_FCN (aeropt) ( int* vulcan, float* aerext, float* aerabs, 
			  float* aersym, float* aerwvn, float* rh, 
			  int* nlyr, float* zd, int* haze, int* maxawvn ); 
  void F77_FCN (aerwvn) ( float* aerwvn, float* aerabs, float* aersym, float* aerext,
			  float* aer_pro, float* aer_dtau, float* aer_gg, 
			  float* aer_ssa, int* nlyr, float* lambda_t, float* zd, 
			  int* nlambda_t, int* maxawvn );



  /* allocate memory for moments even if no aerosol specified */
  output->aer.nmom   = calloc (output->wl.nlambda_t, sizeof(int));
  output->aer.moment = calloc (output->wl.nlambda_t, sizeof(float *));


  if (input.aer.vulcan > 0) {

    /****   Get appropriate aerosol profile and interpolate profile to zd grid.****/

    output->aer.aer_pro  = (float *) calloc(output->atm.nlev, sizeof(float));
    F77_FCN (aerprof) (&output->atm.nlyr, output->atm.zd, output->aer.aer_pro, 
		       &input.aer.seasn, &input.aer.vulcan, &input.aer.visib);

    /****   Get appropriate aerosol optical properties. ****/

    tmp_aerext       = (float *) calloc (maxawvn*output->atm.nlev, sizeof(float));
    tmp_aerabs       = (float *) calloc (maxawvn*output->atm.nlev, sizeof(float));
    tmp_aersym       = (float *) calloc (maxawvn*output->atm.nlev, sizeof(float));
    tmp_aerwvn       = (float *) calloc (maxawvn*output->atm.nlev, sizeof(float));

    F77_FCN (aeropt) (&input.aer.vulcan, tmp_aerext, tmp_aerabs, 
		      tmp_aersym, tmp_aerwvn, output->atm.rh, 
		      &output->atm.nlyr, output->atm.zd, &input.aer.haze, &maxawvn); 

    tmp_aer_dtau       = (float *) calloc (output->wl.nlambda_t*output->atm.nlyr, sizeof(float));
    tmp_aer_gg         = (float *) calloc (output->wl.nlambda_t*output->atm.nlyr, sizeof(float));
    tmp_aer_ssa        = (float *) calloc (output->wl.nlambda_t*output->atm.nlyr, sizeof(float));

    F77_FCN (aerwvn) (tmp_aerwvn, tmp_aerabs, tmp_aersym, tmp_aerext, output->aer.aer_pro, 
		      tmp_aer_dtau, tmp_aer_gg, tmp_aer_ssa, &output->atm.nlyr,
		      output->wl.lambda_t, output->atm.zd, &output->wl.nlambda_t, &maxawvn);

    output->aer.gg   = fortran2c_2D_float_ary (output->wl.nlambda_t, output->atm.nlyr, tmp_aer_gg );
    output->aer.ssa  = fortran2c_2D_float_ary (output->wl.nlambda_t, output->atm.nlyr, tmp_aer_ssa );
    output->aer.dtau = fortran2c_2D_float_ary (output->wl.nlambda_t, output->atm.nlyr, tmp_aer_dtau );
    
    free(tmp_aer_gg);
    free(tmp_aer_ssa);
    free(tmp_aer_dtau);

    free(tmp_aerext);
    free(tmp_aerabs);
    free(tmp_aersym);
    free(tmp_aerwvn);


    /****   Read aerosol optical depth file if specified ****/
    if (strlen(input.aer.filename[FN_AER_TAU]) > 0) {
      status = read_aerosol_profile (input.aer.filename[FN_AER_TAU], output->atm.nlyr,
				     output->wl.nlambda_t, output->atm.zd,
				     output->aer.dtau);
      if (status != 0) {
	fprintf (stderr, "Error %d reading aerosol tau file: %s\n", 
		 status, input.aer.filename[FN_AER_TAU]);
	return status;
      }
    }


    /****   Read aerosol asymmetry parameter file if specified ****/
    if (strlen(input.aer.filename[FN_AER_GG]) > 0) {
      status = read_aerosol_profile (input.aer.filename[FN_AER_GG], output->atm.nlyr,
				     output->wl.nlambda_t, output->atm.zd,
				     output->aer.gg);
      if (status != 0) {
	fprintf (stderr, "Error %d reading aerosol asymmetry parameter file: %s\n", 
		 status, input.aer.filename[FN_AER_GG]);
	return status;
      }
    }


    /****   Read aerosol single scattering albedo file if specified ****/
    if (strlen(input.aer.filename[FN_AER_SSA]) > 0) {
      status = read_aerosol_profile (input.aer.filename[FN_AER_SSA], output->atm.nlyr,
				     output->wl.nlambda_t, output->atm.zd,
				     output->aer.ssa);
      if (status != 0) {
	fprintf (stderr, "Error %d reading aerosol single scattering albedo file: %s\n", 
		 status, input.aer.filename[FN_AER_SSA]);
	return status;
      }
    }




    /****   Read aerosol moments file if specified ****/
    /****   At present: independent of wavelength  ****/
    /****   but this can be easily changed!        ****/
    if (strlen(input.aer.filename[FN_AER_MOM]) > 0) {
      
      status = read_aerosol_moments (input.aer.filename[FN_AER_MOM], 
				     &(output->aer.moment[0]), &(output->aer.nmom[0]));
      if (status != 0) {
	fprintf (stderr, "Error %d reading aerosol moments file: %s\n", 
		 status, input.aer.filename[FN_AER_MOM]);
	return status;
      }
      
      fprintf (stderr, " ... read %d moments from %s\n", 
	       output->aer.nmom[0], input.aer.filename[FN_AER_MOM]);

      /* copy to other wavelengths */
      for (iv=1; iv<output->wl.nlambda_t; iv++)
	for (k=0; k<output->aer.nmom[0]; k++) {
	  output->aer.nmom[iv]      = output->aer.nmom[0];
	  output->aer.moment[iv][k] = output->aer.moment[0][k];
	}
    }


    /****   Read aerosol refractive index   ****/
    if (strlen(input.aer.filename[FN_AER_REF]) > 0) {

      /* read wavelength-dependent aerosol refractive index */
      status = read_3c_file_float (input.aer.filename[FN_AER_REF], 
				   &lref, &re, &im, &nref);
      
      if (status!=0) {
	fprintf (stderr, "Error %d reading aerosol refractive index file\n", status);
	return status;
      }

      /* linear interpolation of the refractive index to the model wavelengths */
      linear = 1;

      output->aer.re_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));
      output->aer.im_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));

      status = arb_wvn (nref, lref, re, 
			output->wl.nlambda_t, output->wl.lambda_t, output->aer.re_t, 
			linear);

      if (status != 0) {
	fprintf (stderr, "arb_wvn reports error %d\n", status);
	return status;
      }

      status = arb_wvn (nref, lref, im, 
			output->wl.nlambda_t, output->wl.lambda_t, output->aer.im_t, linear);

      if (status != 0) {
	fprintf (stderr, "arb_wvn reports error %d\n", status);
	return status;
      }
      
      free(lref); free(re); free(im);

      refrac=1;
    }
    else {
      if (input.aer.re>0) {   /* constant refractive index */
	output->aer.re_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));
	output->aer.im_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));

	for (iv=0; iv<output->wl.nlambda_t; iv++) { 
	  output->aer.re_t[iv] = input.aer.re;
	  output->aer.im_t[iv] = input.aer.im;
	}
	refrac=1;
      }
    }


    /****   Read aerosol size distribution  ****/
    if (strlen(input.aer.filename[FN_AER_SIZ]) > 0) {

      /* read size distribution */
      status = read_2c_file (input.aer.filename[FN_AER_SIZ], 
			     &(output->aer.xsiz), &(output->aer.ysiz), &(output->aer.nsiz));
      
      if (status!=0) {
	fprintf (stderr, "Error %d reading aerosol size distribution file\n", status);
	return status;
      }
      

      if (output->aer.xsiz[0]==0) {
	output->aer.xsiz[0] = 0.1 * output->aer.xsiz[1];
	fprintf (stderr, " ... aerosol size distribution: setting first data point to r=%g\n",
		 output->aer.xsiz[0]);
      }

      sizedist=1;
    }

    if (refrac&&sizedist) {
      fprintf (stderr, " ... now doing aerosol Mie calculation\n");

      mie_inp = calloc (1, sizeof(mie_inp_struct));
      mie_out = calloc (1, sizeof(mie_out_struct));

      mie_out->pmom = calloc (4, sizeof(float *));
      for (ip=0; ip<4; ip++)
	mie_out->pmom[ip] = calloc (nmom, sizeof(float));
      




      /**********************************************************/
      /* The wavelength dependence of the optical thickness     */
      /* follows Mie theory; the absolute value of the optical  */
      /* thickness is defined by its value at this place of the */
      /* code; that is, the value defined by 'visibility' or    */
      /* (much better) by aerosol_set_tau. Later, it may still  */
      /* be scaled with the Angstrom parameters (if anybody     */
      /* wants that.						*/
      /* ??? I guess there are much better ways to do that ???  */
      /**********************************************************/

      for (iv=0; iv<output->wl.nlambda_t; iv++) {

	output->aer.nmom[iv] = nmom;

	mie_inp->anyang      = 0;
	mie_inp->ipolzn      = 0;
	mie_inp->nmom        = nmom;
	mie_inp->momdim      = nmom+1;
	mie_inp->numang      = 0;
	mie_inp->mimcut      = 1.0e-08;
	mie_inp->perfct      = 0;
	mie_inp->prnt[0]     = 0;
	mie_inp->prnt[1]     = 0;
	mie_inp->xmu         = (float *)   calloc (mie_inp->numang+1, sizeof(float));
	mie_inp->s1          = (complex *) calloc (mie_inp->numang+1, sizeof(complex));
	mie_inp->s2          = (complex *) calloc (mie_inp->numang+1, sizeof(complex));


	crefin.re = output->aer.re_t[iv];
	crefin.im = output->aer.im_t[iv];

	status = mie_calc_sizedist (*mie_inp, mie_out, 
				    MIEV0, USER, crefin, 
				    (output->wl.lambda_t[iv])/1000.0, temperature,
				    output->aer.xsiz, output->aer.ysiz, output->aer.nsiz, 
				    &beta, &omega, &g);


	if (status!=0) {
	  fprintf (stderr, "Error %d returned by mie_calc_sizedist()\n", status);
	  return status;
	}

	/* determine maximum non-zero moment */
	kmax=0;
	for (k=1; k<nmom; k++)
	  if (mie_out->pmom[0][k] != 0)
	    kmax=k;
	
	output->aer.nmom[iv]   = kmax+1;
	output->aer.moment[iv] = calloc (kmax+1, sizeof(float));
	
	/* copy moments to final destination */
	for (k=0; k<=kmax; k++)
	  output->aer.moment[iv][k] = mie_out->pmom[0][k]/mie_out->pmom[0][0];
	

	/* extinction at first wavelength; required for scaling the optical thickness */
	if (iv==0)
	  beta0 = beta;


	ext = 0.0;
	  
	/* over-write optical thickness, single-scattering albedo, and asymmetry parameter */
	for (lc=0;lc<output->atm.nlyr;lc++) {
	  output->aer.dtau [iv][lc] *= beta/beta0;
	  output->aer.ssa  [iv][lc]  = omega;
	  output->aer.gg   [iv][lc]  = g;
	  
	  ext += output->aer.dtau[iv][lc];
	}
	

	if (iv==0)
	  fprintf (stderr, "mie  lambda  nmom       tau       ssa       asy\n");
	
	fprintf (stderr, "mie %7.2f  %4d  %g  %g  %g\n",
		 output->wl.lambda_t[iv], kmax, ext, omega, g);

      }

      /* free Mie memory */
      for (ip=0; ip<4; ip++)
	free (mie_out->pmom[ip]);
      free(mie_out->pmom);

      free (mie_inp);
      free (mie_out);
    }


    /**** Set aerosol asymmetry factor ****/
    if (input.aer.set_gg >= -1.0)
      for (lc=0;lc<output->atm.nlyr;lc++)
	for (iv=0;iv<output->wl.nlambda_t;iv++)
	  output->aer.gg[iv][lc] = input.aer.set_gg;
    
    /**** Set aerosol single scattering albedo ****/
    if (input.aer.set_ssa >= 0.0)
      for (lc=0;lc<output->atm.nlyr;lc++)
	for (iv=0;iv<output->wl.nlambda_t;iv++)
	  output->aer.ssa[iv][lc] = input.aer.set_ssa;
    
    /**** Scale aerosol single scattering albedo ****/
    if (input.aer.scale_ssa >= 0.0)
      for (lc=0;lc<output->atm.nlyr;lc++)
	for (iv=0;iv<output->wl.nlambda_t;iv++) {
	  output->aer.ssa[iv][lc] = output->aer.ssa[iv][lc] * input.aer.scale_ssa;
	  if (output->aer.ssa[iv][lc]>1.0) 
	    output->aer.ssa[iv][lc] = 1.0;
	}

    if (input.aer.beta >= 0.0) {
      status = scale_aer_angstrom (output->aer.dtau, output->wl.lambda_t, output->wl.nlambda_t, 
				   output->atm.nlyr, input.aer.alpha, input.aer.beta);

      if (status!=0) {
	fprintf (stderr, "Error %d returned by scale_aer_angstrom()\n", status);
	return status;

      }
    }



  }
  else {
    if ((status = ASCII_calloc_float(&output->aer.dtau, output->wl.nlambda_t, output->atm.nlyr)) != 0 ) return status;
    if ((status = ASCII_calloc_float(&output->aer.gg,   output->wl.nlambda_t, output->atm.nlyr)) != 0 ) return status;
    if ((status = ASCII_calloc_float(&output->aer.ssa,  output->wl.nlambda_t, output->atm.nlyr)) != 0 ) return status;
  }

  return 0;
}



/**************************************************************/
/* Read aerosol profile.                                      */
/**************************************************************/

static int read_aerosol_profile (char *filename, int nlyr, int nlambda_t, 
				 float* zd, float** dtau) 
{
  int rows=0;
  int i=0, j=0;
  int status = 0;
  double *aero_zd=NULL, *aero_tau=NULL;


  /* read altitude and optical depth from aerosol file */
  status = read_2c_file (filename, 
			 &aero_zd, &aero_tau, &rows);

    
  if (status!=0) {
    fprintf (stderr, "Error %d reading aerosol file %s\n", status, filename);
    return status;
  }

  /* check for consistency with atmospheric levels */

  if (rows != nlyr+1) {
    fprintf (stderr, "Aerosol tau file %s: layering differs from atmospheric file\n", 
	     filename); 
    status = 3;
    return status;
  }
  
  for (j=0; j<=nlyr; j++) {
    if (zd[j] != (float) aero_zd[j]) {
      fprintf (stderr, "Aerosol layers differ from atmosphere layers %d %f %f\n", 
	       j, aero_zd[j], zd[j]);
      return -1;
    }
  }

  /* copy optical depth to dtau[][] */
  for (i=0;i<nlambda_t;i++)
    for (j=0;j<nlyr;j++)
      dtau[i][j] = aero_tau[j+1];

  /* free memory */
  free (aero_zd);
  free (aero_tau);

  return 0;
}



/**************************************************************/
/* Read moments of the aerosol phase function; memory for     */
/* float *moment is allocated automatically.                  */
/**************************************************************/

static int read_aerosol_moments (char *filename, 
				 float **moment, int *nmom) 
{
  int i=0;
  int status = 0;
  double *aero_mom=NULL;
  
  /* read moments of the aerosol phase function from file */
  status = read_1c_file (filename, &aero_mom, nmom);
  
  if (status!=0) {
    switch(status)  {
    case ASCIIFILE_NOT_FOUND: 
      fprintf (stderr, "Aerosol tau file %s not found\n", filename);
      return status;
      break;
    case ASCII_NO_MEMORY: 
      fprintf (stderr, "Not enough memory for reading aerosol moments file %s\n",
	       filename);
      return status;
      break;
    default:
      fprintf (stderr, "Error %d reading aerosol file %s\n",  
	       status, filename);
      return status;
      break;
    }
  }

  *moment = calloc (*nmom, sizeof(float));

  /* normalize and copy moments to final destination */

  if (aero_mom[0]<=0) {
    fprintf (stderr, " ... error, first moment cannot be zero or negative\n");
    return -1;
  }


  for (i=0;i<*nmom;i++)
    (*moment)[i] = aero_mom[i] / aero_mom[0];

  /* free memory */
  free (aero_mom);

  return 0;
}




/**************************************************************/
/* Scale the total aerosol optical thickness with Angstrom    */
/* parameters alpha and beta.                                 */
/**************************************************************/

static int scale_aer_angstrom (float **dtau, float *lambda_t, int nlambda_t, int nlyr,
			       float alpha, float beta)
{
  /* Local variables */
  float fact=0, depth=0;
  
  int lc=0, iv=0;
  float ext=0;
  
  
  /* Function Body */
  for (iv=0; iv<nlambda_t; iv++) {
    
    ext = 0.0;
    for (lc=0; lc<nlyr; lc++)
      ext += dtau[iv][lc];
    
    depth = beta * pow(lambda_t[iv]/1000.0, -alpha);
    fact = depth / ext;
    
    for (lc=0; lc<nlyr; lc++)
      dtau[iv][lc] *= fact;
  }
  
  return 0;
} 







