/************************************************************************/
/*                                                                      */
/* Copyright (C) 1997 Arve Kylling                                      */
/*                                                                      */
/* This program is free software; you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation; either version 1, or (at your option)  */
/* any later version.                                                   */
/*                                                                      */
/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY of FITNESS FOR A PARTICULAR PURPOSE. See the         */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* To obtain a copy of the GNU General Public License write to the      */
/* Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,   */
/* USA.                                                                 */
/*                                                                      */
/************************************************************************/

#include <stdio.h>							    
#include <stdlib.h>
#include "uvspec.h"
#include "../src_c/ascii.h"



/************************************/
/* prototypes of internal functions */
/************************************/

static int read_albedo(char* albedo_filename, 
		       atm_inp_struct *atm_inp, 
		       atm_out_struct *atm_out);



/**************************************************************/
/* Setup surface albedo.                                      */
/**************************************************************/

int setup_albedo (input_struct input, output_struct *output)
{
  int status=0, iv=0, linear=0;

  output->atm.albedo_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));

  output->rpv.rho0_t   = (float *) calloc (output->wl.nlambda_t, sizeof(float));
  output->rpv.k_t      = (float *) calloc (output->wl.nlambda_t, sizeof(float));
  output->rpv.theta_t  = (float *) calloc (output->wl.nlambda_t, sizeof(float));


  if (input.atm.albedo < 0.0) {
    status = read_albedo(input.filename[FN_ALBEDO], &input.atm, &output->atm);
    if (status != 0) {
      fprintf (stderr, "Error reading albedo file: %s\n", input.filename[FN_ALBEDO]);
      return status;
    }
    
    /**** Interpolate albedo to transmission grid ****/
    
    linear = 1;  /**** Use linear interpolation for the albedo, otherwise 
		       negative albedos or albedos larger than 1 may occur, 
		       I should know:-) AKy, ****/ 

    status = arb_wvn (output->atm.file_nlambda_albedo, output->atm.file_lambda_albedo,
		      output->atm.file_albedo, 
		      output->wl.nlambda_t, output->wl.lambda_t, output->atm.albedo_t, 
		      linear);
    if (status != 0) {
      fprintf (stderr, "arb_wvn reports error %d\n", status);
      return status;
    }
    
    /**** Check interpolated albedo values ****/

    for (iv=0; iv<output->wl.nlambda_t; iv++) {
      if (output->atm.albedo_t[iv] > 1.0 || output->atm.albedo_t[iv] < 0.0)  {
	fprintf (stderr, "Albedo interpolated from %s out of range:\n", input.filename[FN_ALBEDO]);
	fprintf (stderr, "wavelength: %f, albedo %f\n", 
		 output->wl.lambda_t[iv], output->atm.albedo_t[iv]);
	fprintf (stderr, "Modify albedo file %s\n", input.filename[FN_ALBEDO]);
	return -1;
      } 
    }
  }
  else
    for (iv=0;iv<output->wl.nlambda_t;iv++)
      output->atm.albedo_t[iv] = input.atm.albedo;    
  

  /* read BRDF */
  if (strlen(input.filename[FN_RPV]) > 0) {
    
    status = read_4c_file_float (input.filename[FN_RPV],
				 &(output->rpv.lambda),
				 &(output->rpv.rho0),
				 &(output->rpv.k),
				 &(output->rpv.theta),
				 &(output->rpv.n));
    
    if (status!=0) {
      fprintf (stderr, "error %d reading %s\n", status, input.filename[FN_RPV]);
      return status;
    }

    fprintf (stderr, " ... read rpv data from %g to %g\n",
	     output->rpv.lambda[0], 
	     output->rpv.lambda[output->rpv.n-1]); 


    linear=1;   /* linear interpolation */

    /* interpolate to internal RTE grid */

    fprintf (stderr, " ... interpolating rho0 to %g - %g\n, \n",
	     output->wl.lambda_t[0], output->wl.lambda_t[output->wl.nlambda_t-1]);
    status = arb_wvn (output->rpv.n, output->rpv.lambda, output->rpv.rho0, 
		      output->wl.nlambda_t, output->wl.lambda_t, output->rpv.rho0_t, 
		      linear);

    if (status != 0) {
      fprintf (stderr, "arb_wvn (rho0) reports error %d\n", status);
      return status;
    }

    status = arb_wvn (output->rpv.n, output->rpv.lambda, output->rpv.k, 
		      output->wl.nlambda_t, output->wl.lambda_t, output->rpv.k_t, 
		      linear);

    if (status != 0) {
      fprintf (stderr, "arb_wvn (k) reports error %d\n", status);
      return status;
    }

    status = arb_wvn (output->rpv.n, output->rpv.lambda, output->rpv.theta, 
		      output->wl.nlambda_t, output->wl.lambda_t, output->rpv.theta_t, 
		      linear);

    if (status != 0) {
      fprintf (stderr, "arb_wvn (theta) reports error %d\n", status);
      return status;
    }
  }
  else {
    for (iv=0;iv<output->wl.nlambda_t;iv++) {
      output->rpv.rho0_t  [iv] = input.rpv.rho0;    
      output->rpv.k_t     [iv] = input.rpv.k;    
      output->rpv.theta_t [iv] = input.rpv.theta;    
    }
  }

  return 0;
}





/**************************************************************/
/* Read albedo from file.                                     */
/**************************************************************/

static int read_albedo (char* albedo_filename,  
			atm_inp_struct *atm_inp, atm_out_struct *atm_out) 
{
  int status=0;
  
  status = read_2c_file_float (albedo_filename, 
			       &(atm_out->file_lambda_albedo),
			       &(atm_out->file_albedo),
			       &(atm_out->file_nlambda_albedo));

  if (status!=0) {
    fprintf (stderr, "error %d reading %s\n", status, albedo_filename);
    return status;
  }

  return 0;
}

