/************************************************************************/
/*                                                                      */
/* Copyright (C) 1997 Arve Kylling                                      */
/*                                                                      */
/* This program is free software; you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation; either version 1, or (at your option)  */
/* any later version.                                                   */
/*                                                                      */
/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY of FITNESS FOR A PARTICULAR PURPOSE. See the         */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* To obtain a copy of the GNU General Public License write to the      */
/* Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,   */
/* USA.                                                                 */
/*                                                                      */
/************************************************************************/

#include <stdio.h>							    
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include <ckdfu.h>
#include "uvspec.h"
#include "../src_c/ascii.h"
#include "../src_c/numeric.h"

#define EPSILON 1E-6

#ifndef PI
#define PI 3.14159265358979323846264338327
#endif

/* prototypes of internal functions */
static int cnvlv (double *x_spec, float  *y_spec, int n_spec,
		  double *x_slit, double *y_slit, int n_slit);


/***************************************************************/
/* Setup the wavelength grid for the transmission calculation. */
/***************************************************************/

int setup_wlgrid (input_struct input, output_struct *output)
{     
  int status=0, iv=0, rows=0, min_columns=0, max_columns=0;
  float **wl_data=NULL;

  if (input.ck_scheme == CK_NONE) {   /* no correlated-k */
    if ( strlen(input.filename[FN_WLTRANS]) > 0 ) {
      
      status = ASCII_file2float (input.filename[FN_WLTRANS], 
				 &rows, &max_columns, &min_columns, 
				 &wl_data);
      if (status!=0) {
	fprintf (stderr, " ... error %d opening file %s\n", 
		 status, input.filename[FN_WLTRANS]);
	return status;
      }
      
      if (max_columns!=min_columns) {
	fprintf (stderr, " !! ATTENTION !! Inconsistent number of columns\n");
	fprintf (stderr, "     min = %d, max =%d\n", min_columns, max_columns);
      }  
      
      if (min_columns!=1) {
	fprintf (stderr, " ... ending, too few columns\n");
	return -1;
      }
      
      output->wl.nlambda_t = rows;
      output->wl.lambda_t  = ASCII_column_float(wl_data, output->wl.nlambda_t, 0 );
      
      status = ASCII_free_float(wl_data, rows);
    }
    else 
      status = set_transmittance_wl_grid(&input.wl, &output->wl);


    /* allocate memory for the wavenumber arrays; they are currently */
    /* only required for correlated-k, but to avoid stupid errors    */
    /* memory is allocated anyway.                                   */
    output->wl.wvnmlo_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));
    output->wl.wvnmhi_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));

    /* here, a wavelength interval of 1 cm-1 is chosen to get */
    /* the emittance per cm-1                                 */       
    for (iv=0; iv<output->wl.nlambda_t; iv++) {
      output->wl.wvnmlo_t[iv] = 1.0E7 / output->wl.lambda_t[iv] - 0.5;
      output->wl.wvnmhi_t[iv] = output->wl.wvnmlo_t[iv] + 1.0;
    }

  }
  else { /* correlated-k */
    switch(input.ck_scheme) {
    case CK_KATO:

      /* read tables */
      status = kato_readtables (&(output->ck), input.filename[FN_PATH],
				input.rte.mc.filename[FN_MC_PHOTONS]);

      if (status!=0) {
	fprintf (stderr, " ... error %d returned by kato_readtables()\n", status); 
	return status;
      }
      
      break;

    case CK_FU:

      /* read tables */
      status = fu_readtables (&(output->ck), input.filename[FN_PATH],
				input.rte.mc.filename[FN_MC_PHOTONS]);

      if (status!=0) {
	fprintf (stderr, " ... error %d returned by fu_readtables()\n", status); 
	return status;
      }
      
      break;

    case CK_AVHRR_KRATZ:

      /* read tables */
      status = avhrr_kratz_readtables (&(output->ck), input.filename[FN_PATH]);

      if (status!=0) {
	fprintf (stderr, " ... error %d returned by avhrr_kratz_readtables()\n", status); 
	return status;
      }
      
      break;

    default:
      fprintf (stderr, " ... error: unsupported correlated-k scheme\n");
      return -1;
    }


    /* copy center wavelengths to transmittance grid */
    /* and set wavenumber intervals                  */

    output->wl.nlambda_t = output->ck.n_wvl;
    output->wl.lambda_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));
    output->wl.wvnmlo_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));
    output->wl.wvnmhi_t = (float *) calloc (output->wl.nlambda_t, sizeof(float));

    for (iv=0; iv<output->wl.nlambda_t; iv++) {
      output->wl.lambda_t[iv] = output->ck.wvlc[iv+1];
      output->wl.wvnmlo_t[iv] = 1.0E7 / output->ck.wvl2[iv+1];
      output->wl.wvnmhi_t[iv] = 1.0E7 / output->ck.wvl1[iv+1];
    }
  }
    
  return 0;
}  






/**************************************************************/
/* Interpolate from one wavelength grid to another, either    */
/* with natural cubic splines (linear=0) or linear (linear=1) */ 
/**************************************************************/

int arb_wvn( int n_old_x, float *old_x, float *old_y,  
	     int n_new_x, float *new_x, float *new_y, int linear)
{
  int i=0, status=0;
  double *a0=NULL, *a1=NULL, *a2=NULL, *a3=NULL;
  double *x=NULL, *y=NULL;
  double ynew=0.0;

  x = (double *) calloc (n_old_x, sizeof(double));
  y = (double *) calloc (n_old_x, sizeof(double));

  for (i=0;i<n_old_x;i++) {
    x[i] = (double) old_x[i];
    y[i] = (double) old_y[i];
  }

  if (linear) {
    status = linear_coeffc (x, y, n_old_x, &a0, &a1, &a2, &a3);
    if (status!=0)  {
      fprintf (stderr, "arb_wvn: sorry cannot do linear interpolation\n");
      fprintf (stderr, "linear_coeffc() returned status %d\n", status);
      return status;
    }
  }
  else {
    status = spline_coeffc (x, y, n_old_x, &a0, &a1, &a2, &a3);
    if (status!=0)  {
      fprintf (stderr, "arb_wvn: sorry cannot do spline interpolation\n");
      fprintf (stderr, "spline_coeffc() returned status %d\n", status);
      return status;
    }
  }

  for (i=0; i<n_new_x; i++)  {
    status = calc_splined_value ((double) new_x[i], &ynew, x, n_old_x, a0, a1, a2, a3);

    /* this check is new in 0.99-alpha-6; so far, no error */
    /* was reported if a value could not be interpolated   */
    if (status!=0) {
      fprintf (stderr, "error %d returned by calc_splined_value(%g)\n",
	       status, new_x[i]);
      return status;
    }

    new_y[i] = (float) ynew;
  }

  free(a0); free(a1); free(a2); free(a3);
  free(x); free(y);

  return 0;
}




/**************************************************************/
/* Define the internal wavelength grid for the transmittance  */
/* calculation.                                               */
/**************************************************************/

int set_transmittance_wl_grid(wl_inp_struct *wl_inp, wl_out_struct *wl_out)
{
  int iv=0;
  float lambda=0.0, wvn_step_t=0.0;
  
  /* Calculate number of wavelengths needed for transmittance calculation */

  lambda = wl_inp->wl_start;
  iv = 1;
  while (lambda < wl_inp->wl_end) {
    if ( lambda < 350.0 )
      wvn_step_t  = 0.5;
    else
      wvn_step_t  = 1.0;

    lambda += wvn_step_t;
    iv++;
  }

  wl_out->nlambda_t = iv;
  wl_out->lambda_t = (float *) calloc (wl_out->nlambda_t, sizeof(float));


  /* Set wavelengths for transmittance calculation */
  lambda = wl_inp->wl_start;
  for (iv=0; iv<wl_out->nlambda_t; iv++) {
    wl_out->lambda_t[iv] = lambda;
    if ( lambda < 350.0 )
      wvn_step_t  = 0.5;
    else
      wvn_step_t  = 1.0;

    lambda += wvn_step_t;
  }

  return 0;
}



/**************************************************************/
/* Convolve the spectrum with a user-defined slit function    */
/**************************************************************/

int convolve (input_struct input, output_struct *output)
{
  int is=0, js=0, ks=0, iv=0, iu=0, j=0, lev=0;
  int status=0;
  double *x_data=NULL;

  int n_slit=0;
  double *x_slit=NULL, *y_slit=NULL;


  /* read slit function from file */
  status = read_2c_file (input.filename[FN_SLITFUNCTION], &x_slit, &y_slit, &n_slit); 
  if (status!=0)  { 
    fprintf (stderr, " ... error %d reading file %s\n", status, input.filename[FN_SLITFUNCTION]); 
    return status; 
  } 

  x_data = calloc (output->wl.nlambda_h, sizeof(double)); 

  for (iv=0; iv<output->wl.nlambda_h; iv++)
    x_data[iv] = (double) output->wl.lambda_h[iv];
  
  if (input.rte.solver == POLRADTRAN ) {
    for (lev=0; lev<input.atm.nzout; lev++) {

      /* Convolve up_flux and down_flux*/
      for (is=0; is<input.rte.pol_nstokes; is++) {
	status = cnvlv (x_data, output->up_flux[lev][is], output->wl.nlambda_h,
			x_slit, y_slit, n_slit);
	if (status!=0)  { 
	  fprintf (stderr, " ... error %d convoluting up_flux\n", status); 
	  return status; 
	} 

	status = cnvlv (x_data, output->down_flux[lev][is], output->wl.nlambda_h,
			x_slit, y_slit, n_slit);
	if (status!=0)  { 
	  fprintf (stderr, " ... error %d convoluting down_flux\n", status); 
	  return status; 
	} 

	/* Convolve up_rad and down_rad*/
	for (j=0; j<input.rte.nphi; j++) {
	  for (iu=0; iu<input.rte.nstr/2; iu++) {
	    status = cnvlv (x_data, output->down_rad[lev][j][iu][is], output->wl.nlambda_h,
			    x_slit, y_slit, n_slit);
	    if (status!=0)  { 
	      fprintf (stderr, " ... error %d convoluting down_rad\n", status); 
	      return status; 
	    } 

	    status = cnvlv (x_data, output->up_rad[lev][j][iu][is], output->wl.nlambda_h,
			    x_slit, y_slit, n_slit);
	    if (status!=0)  { 
	      fprintf (stderr, " ... error %d convoluting up_rad\n", status); 
	      return status; 
	    } 
	  }
	}
      }
    }
  }
  else {

    for (lev=0; lev<input.atm.nzout; lev++) {

      /* Convolve rfldir, rfldn, flup and uavg for all cases */

      /* rfldir */
      status = cnvlv (x_data, output->rfldir[lev], output->wl.nlambda_h,
		      x_slit, y_slit, n_slit);

      if (status!=0)  { 
	fprintf (stderr, " ... error %d convoluting rfldir\n", status); 
	return status; 
      } 
      
      /* rfldn */
      status = cnvlv (x_data, output->rfldn[lev], output->wl.nlambda_h,
		      x_slit, y_slit, n_slit);
      if (status!=0)  { 
	fprintf (stderr, " ... error %d convoluting rfldn\n", status); 
	return status; 
      } 
      
      /* flup */
      status = cnvlv (x_data, output->flup[lev], output->wl.nlambda_h,
		      x_slit, y_slit, n_slit);
      if (status!=0)  { 
	fprintf (stderr, " ... error %d convoluting flup\n", status); 
	return status; 
      } 
      
      /* uavg */
      status = cnvlv (x_data, output->uavg[lev], output->wl.nlambda_h,
		      x_slit, y_slit, n_slit);
      if (status!=0)  { 
	fprintf (stderr, " ... error %d convoluting uavg\n", status); 
	return status; 
      } 
      
      
      /* 3D fields */
      if (input.rte.mc.passback3D)
	for (is=0; is<input.rte.mc.Nx_sample; is++)
	  for (js=0; js<input.rte.mc.Ny_sample; js++) {

	    /* rfldir3d */
	    status = cnvlv (x_data, output->rfldir3d[lev][is][js], output->wl.nlambda_h,
			    x_slit, y_slit, n_slit);
	    if (status!=0)  { 
	      fprintf (stderr, " ... error %d convoluting rfldir3d\n", status); 
	      return status; 
	    } 

	    /* rfldn3d */
	    status = cnvlv (x_data, output->rfldn3d[lev][is][js], output->wl.nlambda_h,
			    x_slit, y_slit, n_slit);
	    if (status!=0)  { 
	      fprintf (stderr, " ... error %d convoluting rfldir3d\n", status); 
	      return status; 
	    } 

	    /* flup3d */
	    status = cnvlv (x_data, output->flup3d[lev][is][js], output->wl.nlambda_h,
			    x_slit, y_slit, n_slit);
	    if (status!=0)  { 
	      fprintf (stderr, " ... error %d convoluting rfldir3d\n", status); 
	      return status; 
	    } 
	  }
	

      if (input.rte.solver != TWOSTR) {
	
	/* Convolve uavgso, uavgdn, uavgup */
	
	/* uavgso */
	status = cnvlv (x_data, output->uavgso[lev], output->wl.nlambda_h,
			x_slit, y_slit, n_slit);
	if (status!=0)  { 
	  fprintf (stderr, " ... error %d convoluting uavgso\n", status); 
	  return status; 
	} 
	
	/* uavgdn */
	status = cnvlv (x_data, output->uavgdn[lev], output->wl.nlambda_h,
			x_slit, y_slit, n_slit);
	if (status!=0)  { 
	  fprintf (stderr, " ... error %d convoluting uavgdn\n", status); 
	  return status; 
	} 
	
	/* uavgup */
	status = cnvlv (x_data, output->uavgup[lev], output->wl.nlambda_h,
			x_slit, y_slit, n_slit);
	if (status!=0)  { 
	  fprintf (stderr, " ... error %d convoluting uavgup\n", status); 
	  return status; 
	} 

	
	if ( input.rte.numu > 0 ) {
	  
	  /* Convolve u0u */
	  for (iu=0; iu<input.rte.maxumu; iu++) {
	    status = cnvlv (x_data, output->u0u[lev][iu], output->wl.nlambda_h,
			    x_slit, y_slit, n_slit);
	    if (status!=0)  { 
	      fprintf (stderr, " ... error %d convoluting u0u\n", status); 
	      return status; 
	    } 
	  }
	  
	  if ( output->print.phi > 0 ) {
	    
	    /* Convolve uu */
	    for (j=0; j<input.rte.nphi; j++) {
	      for (iu=0; iu<input.rte.maxumu; iu++) {
		status = cnvlv (x_data, output->uu[lev][j][iu], output->wl.nlambda_h,
				x_slit, y_slit, n_slit);
		if (status!=0)  { 
		  fprintf (stderr, " ... error %d convoluting u0u\n", status); 
		  return status; 
		} 
	      }
	    }
	  }
	}
      }
    }

    /* 3D absorption */
    if (input.rte.mc.passback3D && input.rte.mc.absorption)

      for (ks=0; ks<input.rte.mc.Nz_cloud; ks++)
	for (is=0; is<input.rte.mc.Nx_cloud; is++)
	  for (js=0; js<input.rte.mc.Ny_cloud; js++) {
	    
	    status = cnvlv (x_data, output->abs3d[ks][is][js], output->wl.nlambda_h,
			    x_slit, y_slit, n_slit);
	    if (status!=0)  { 
	      fprintf (stderr, " ... error %d convoluting abs3d\n", status); 
	      return status; 
	    } 
	  }
  }


  free(x_data);  

  return 0;
}



/**************************************************************/
/* Convolve a spectrum with a slit function.                  */
/* Internal function, used by convolve().                     */
/**************************************************************/

static int cnvlv (double *x_spec, float  *y_spec, int n_spec,
		  double *x_slit, double *y_slit, int n_slit) 
{
  int iv=0, status=0;
  double *y_data = calloc (n_spec, sizeof(double));
  double *tmp=NULL;

  for (iv=0; iv<n_spec; iv++) 
    y_data[iv] = (double) y_spec[iv];
  
  status = int_convolute (x_spec, y_data, n_spec, x_slit, y_slit, n_slit, &tmp);
  if (status!=0) 
    return status; 
  
  for (iv=0; iv<n_spec; iv++) 
    y_spec[iv] = (float) tmp[iv];
  
  free(y_data);
  free(tmp);

  return 0;
}


/**************************************************************/
/* Interpolate the hires spectrum to a user-defined           */
/* wavelength grid.                                           */
/**************************************************************/

int spline_interpolate(input_struct input, output_struct *output)
{
  int is=0, js=0, ks=0, iu=0, iv=0, j=0, lev=0;
  int status=0;
  double *x_user=NULL;
  int linear=0;

  if (strlen(input.filename[FN_SPLINE]) > 0) {

    /* read file with user x values */
    status = read_1c_file (input.filename[FN_SPLINE], &x_user, &output->wl.nlambda_s);
    if (status!=0)  {
      fprintf (stderr, "spline_interpolate: error %d reading file %s\n", 
	       status, input.filename[FN_SPLINE]);
      return (status);
    }

    output->wl.lambda_s = (float *) calloc (output->wl.nlambda_s, sizeof(float));

    for (iv=0; iv<output->wl.nlambda_s; iv++)
      output->wl.lambda_s[iv] = (float) x_user[iv];

    /* Check wavelength ranges */
    status=0;
    if (input.wl.wl_start > output->wl.lambda_s[0])  {
      fprintf (stderr, "Spline lambda %f is smaller than wvn lambda %f\n",
	       output->wl.lambda_s[0], input.wl.wl_start);
      status--;
    }

    if (input.wl.wl_end < output->wl.lambda_s[output->wl.nlambda_s-1])  {
      fprintf (stderr, "Spline lambda %f is greater than wvn lambda %f\n",
	       output->wl.lambda_s[output->wl.nlambda_s-1], input.wl.wl_end);
      status--;
    }

    if (status!=0) 
      return status;
  }
  else {
    output->wl.nlambda_s = 
      (int) ((input.spline_lambda_1-input.spline_lambda_0)/input.spline_lambda_step+1);

    output->wl.lambda_s = (float *) calloc (output->wl.nlambda_s, sizeof(float));

    for (iv=0;iv<output->wl.nlambda_s;iv++)
      output->wl.lambda_s[iv] = input.spline_lambda_0 + (float) iv * input.spline_lambda_step;
  }
  

  /* Spline interpolate to given wavelengths */
  if (input.rte.solver == POLRADTRAN ) {
    for (lev=0; lev<input.atm.nzout; lev++) {
      /* Spline interpolate up_flux and down_flux*/
      for (is=0;is<input.rte.pol_nstokes;is++) {
	status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->up_flux[lev][is],  
			 output->wl.nlambda_s, output->wl.lambda_s, output->up_flux[lev][is], linear);
	if (status!=0) {
	  fprintf (stderr, " ... error %d interpolating up_flux\n", status);
	  return status;
	}
	status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->down_flux[lev][is],  
			 output->wl.nlambda_s, output->wl.lambda_s, output->down_flux[lev][is], linear);
	if (status!=0) {
	  fprintf (stderr, " ... error %d interpolating down_flux\n", status);
	  return status;
	}
	/* Spline interpolate up_rad and down_rad */
	for (j=0;j<input.rte.nphi;j++) {
	  for (iu=0;iu<input.rte.nstr/2;iu++) {
	    status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->down_rad[lev][j][iu][is],  
			     output->wl.nlambda_s, output->wl.lambda_s, output->down_rad[lev][j][iu][is],
			     linear);
	    if (status!=0) {
	      fprintf (stderr, " ... error %d interpolating down_rad\n", status);
	      return status;
	    }
	    status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->up_rad[lev][j][iu][is],  
			     output->wl.nlambda_s, output->wl.lambda_s, output->up_rad[lev][j][iu][is],
			     linear);
	    if (status!=0) {
	      fprintf (stderr, " ... error %d interpolating up_rad\n", status);
	      return status;
	    }
	  }
	}
      }
    }
  }
  else {
    for (lev=0; lev<input.atm.nzout; lev++) {

      /* Spline interpolate rfldir, rfldn, flup and uavg for all cases */
	  
      /* rfldir */
      status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->rfldir[lev],  
		       output->wl.nlambda_s, output->wl.lambda_s, output->rfldir[lev], linear);
      if (status!=0) {
	fprintf (stderr, " ... error %d interpolating rfldir\n", status);
	return status;
      }
      
      /* rfldn */
      status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->rfldn[lev],  
		       output->wl.nlambda_s, output->wl.lambda_s, output->rfldn[lev], linear);
      if (status!=0) {
	fprintf (stderr, " ... error %d interpolating rfldn\n", status);
	return status;
      }
      
      /* flup */
      status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->flup[lev],  
		       output->wl.nlambda_s, output->wl.lambda_s, output->flup[lev], linear);
      if (status!=0) {
	fprintf (stderr, " ... error %d interpolating flup\n", status);
	return status;
      }
      
      /* uavg */
      status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->uavg[lev],  
		       output->wl.nlambda_s, output->wl.lambda_s, output->uavg[lev], linear);
      if (status!=0) {
	fprintf (stderr, " ... error %d interpolating uavg\n", status);
	return status;
      }
      

      /* 3D fields */
      if (input.rte.mc.passback3D)
	for (is=0; is<input.rte.mc.Nx_sample; is++)
	  for (js=0; js<input.rte.mc.Ny_sample; js++) {
	    status = arb_wvn(output->wl.nlambda_t, output->wl.lambda_h, output->rfldir3d[lev][is][js], 
			     output->wl.nlambda_h, output->wl.lambda_s, output->rfldir3d[lev][is][js], 
			     linear);
	    if (status!=0) {
	      fprintf (stderr, " ... error %d interpolating rfldir3d\n", status);
	      return status;
	    }
	    
	    status = arb_wvn(output->wl.nlambda_t, output->wl.lambda_h, output->rfldn3d[lev][is][js], 
			     output->wl.nlambda_h, output->wl.lambda_s, output->rfldn3d[lev][is][js], 
			     linear);
	    if (status!=0) {
	      fprintf (stderr, " ... error %d interpolating rfldn3d\n", status);
	      return status;
	    }
	    
	    status = arb_wvn(output->wl.nlambda_t, output->wl.lambda_h, output->flup3d[lev][is][js], 
			     output->wl.nlambda_h, output->wl.lambda_s, output->flup3d[lev][is][js], 
			     linear);
	    if (status!=0) {
	      fprintf (stderr, " ... error %d interpolating flup3d\n", status);
	      return status;
	    }

	  }
      
      if (input.rte.solver != TWOSTR) {
	
	/* Spline interpolate uavgso, uavgdn, uavgup */
	
	/* uavgso */
	status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->uavgso[lev],  
			 output->wl.nlambda_s, output->wl.lambda_s, output->uavgso[lev], linear);
	if (status!=0) {
	  fprintf (stderr, " ... error %d interpolating uavgso\n", status);
	  return status;
	}
	
	/* uavgdn */
	status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->uavgdn[lev],  
			 output->wl.nlambda_s, output->wl.lambda_s, output->uavgdn[lev], linear);
	if (status!=0) {
	  fprintf (stderr, " ... error %d interpolating uavgdn\n", status);
	  return status;
	}
	
	/* uavgup */
	status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->uavgup[lev],  
			 output->wl.nlambda_s, output->wl.lambda_s, output->uavgup[lev], linear);
	if (status!=0) {
	  fprintf (stderr, " ... error %d interpolating uavgup\n", status);
	  return status;
	}
	
	if (input.rte.numu > 0) {
	  
	  /* Spline interpolate u0u */
	  for (iu=0;iu<input.rte.maxumu;iu++) {
	    status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->u0u[lev][iu],  
			     output->wl.nlambda_s, output->wl.lambda_s, output->u0u[lev][iu], linear);
	    if (status!=0) {
	      fprintf (stderr, " ... error %d interpolating u0u\n", status);
	      return status;
	    }
	  }
	  
	  if ( output->print.phi > 0 ) {
	    
	    /* Spline interpolate uu */
	    for (j=0;j<input.rte.nphi;j++) {
	      for (iu=0;iu<input.rte.maxumu;iu++) {
		status = arb_wvn(output->wl.nlambda_h, output->wl.lambda_h, output->uu[lev][j][iu],  
				 output->wl.nlambda_s, output->wl.lambda_s, output->uu[lev][j][iu], linear);
		if (status!=0) {
		  fprintf (stderr, " ... error %d interpolating uu\n", status);
		  return status;
		}
	      }
	    }
	  }
	}
      }
    }


    if (input.rte.mc.passback3D && input.rte.mc.absorption)
      for (ks=0; ks<input.rte.mc.Nz_cloud; ks++)
	for (is=0; is<input.rte.mc.Nx_cloud; is++)
	  for (js=0; js<input.rte.mc.Ny_cloud; js++) {

	    
	    status = arb_wvn(output->wl.nlambda_t, output->wl.lambda_h, output->abs3d[ks][is][js], 
			     output->wl.nlambda_h, output->wl.lambda_s, output->abs3d[ks][is][js], 
			     linear);
	    if (status!=0) {
	      fprintf (stderr, " ... error %d interpolating abs3d\n", status);
	      return status;
	    }
	  }

  }
  

  free(output->wl.lambda_h);
  output->wl.nlambda_h = output->wl.nlambda_s;
  output->wl.lambda_h = (float *) calloc (output->wl.nlambda_s, sizeof(float));
  for (iv=0;iv<output->wl.nlambda_h;iv++)
    output->wl.lambda_h[iv] = (float) output->wl.lambda_s[iv];

  return status;
  
}





/**************************************************************/
/* Interpolate the transmittance from the internal grid to    */
/* high resolution grid.                                      */
/**************************************************************/

int interpolate_transmittance (input_struct input, output_struct *output)
{
  int lu=0, is=0, js=0, ks=0, iv=0, ivh=0, iu=0, j=0;
  int linear=0, status=0;


  /* If only one wavelength is required, or in correlated-k mode, */
  /* arrays are not interpolated but copied                       */
  
  if (output->wl.nlambda_t == 1 || input.ck_scheme!=CK_NONE) {
    
    for (ivh=0; ivh<output->wl.nlambda_h; ivh++) {
      
      if (output->wl.nlambda_t == 1)
	iv = ivh;
      else 
	iv = output->wl.map_e2h[ivh];
      
      for (lu=0; lu<input.atm.nzout; lu++) {
	if (input.rte.solver == POLRADTRAN) {

	  for (is=0;is<input.rte.pol_nstokes;is++) {
	    output->up_flux   [lu][is][ivh] = output->up_flux_t   [lu][is][iv];
	    output->down_flux [lu][is][ivh] = output->down_flux_t [lu][is][iv];
	  }

	  for (j=0;j<input.rte.nphi;j++)
	    for (iu=0;iu<input.rte.nstr/2;iu++)
	      for (is=0;is<input.rte.pol_nstokes;is++) {
		output->down_rad [lu][j][iu][is][iv] = output->down_rad_t [lu][j][iu][is][iv];
		output->up_rad   [lu][j][iu][is][iv] = output->up_rad_t   [lu][j][iu][is][iv];
	      } 
	  
	}

	else {
	  output->flup   [lu][ivh] = output->flup_t   [lu][iv];
	  output->rfldir [lu][ivh] = output->rfldir_t [lu][iv];
	  output->rfldn  [lu][ivh] = output->rfldn_t  [lu][iv];
	  output->uavg   [lu][ivh] = output->uavg_t   [lu][iv];
	  output->uavgdn [lu][ivh] = output->uavgdn_t [lu][iv];
	  output->uavgso [lu][ivh] = output->uavgso_t [lu][iv];
	  output->uavgup [lu][ivh] = output->uavgup_t [lu][iv];


	  /* radiances */
	  for (iu=0; iu<input.rte.maxumu; iu++) {
	    output->u0u[lu][iu][iv] = output->u0u_t[lu][iu][iv];

	    for (j=0; j<input.rte.nphi; j++)
	      output->uu[lu][j][iu][iv] = output->uu_t[lu][j][iu][iv];
	  }

	  /* 3D fields */
	  if (input.rte.mc.passback3D)
	    for (is=0; is<input.rte.mc.Nx_sample; is++)
	      for (js=0; js<input.rte.mc.Ny_sample; js++) {
		output->rfldir3d [lu][is][js][ivh] = output->rfldir3d_t [lu][is][js][iv];
		output->rfldn3d  [lu][is][js][ivh] = output->rfldn3d_t  [lu][is][js][iv];
		output->flup3d   [lu][is][js][ivh] = output->flup3d_t   [lu][is][js][iv];
	      }
	}
      }
      
      /* absorption is defined on the 3D cloud grid, not on the user-defined grid */
      if (input.rte.mc.passback3D && input.rte.mc.absorption)
	for (ks=0; ks<input.rte.mc.Nz_cloud; ks++)
	  for (is=0; is<input.rte.mc.Nx_cloud; is++)
	    for (js=0; js<input.rte.mc.Ny_cloud; js++)
	      output->abs3d [ks][is][js][ivh] = output->abs3d_t [ks][is][js][iv];
      
      
      output->sza [ivh] = output->atm.sza_t [iv];
    }
  } 
  else {   /* interpolation to differemt wavelengths required */
    for (lu=0; lu<input.atm.nzout; lu++) {

      if (input.rte.solver == POLRADTRAN) {
	for (is=0;is<input.rte.pol_nstokes;is++) {
	  status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->down_flux_t[lu][is], 
			    output->wl.nlambda_h, output->wl.lambda_h, output->down_flux[lu][is], linear);
	  status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->up_flux_t[lu][is], 
			    output->wl.nlambda_h, output->wl.lambda_h, output->up_flux[lu][is], linear);
	}
      }
      else {
	status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->flup_t[lu], 
			  output->wl.nlambda_h, output->wl.lambda_h, output->flup[lu], linear);
	status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->rfldir_t[lu], 
			  output->wl.nlambda_h, output->wl.lambda_h, output->rfldir[lu], linear);
	status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->rfldn_t[lu], 
			  output->wl.nlambda_h, output->wl.lambda_h, output->rfldn[lu], linear);
	status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->uavg_t[lu], 
			  output->wl.nlambda_h, output->wl.lambda_h, output->uavg[lu], linear);
	status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->uavgdn_t[lu], 
			  output->wl.nlambda_h, output->wl.lambda_h, output->uavgdn[lu], linear);
	status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->uavgso_t[lu], 
			  output->wl.nlambda_h, output->wl.lambda_h, output->uavgso[lu], linear);
	status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->uavgup_t[lu], 
			  output->wl.nlambda_h, output->wl.lambda_h, output->uavgup[lu], linear);
	
	/* 3D fields */
	if (input.rte.mc.passback3D)
	  for (is=0; is<input.rte.mc.Nx_sample; is++)
	    for (js=0; js<input.rte.mc.Ny_sample; js++) {
	      status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->rfldir3d_t[lu][is][js], 
				output->wl.nlambda_h, output->wl.lambda_h, output->rfldir3d[lu][is][js], linear);

	      status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->rfldn3d_t[lu][is][js], 
				output->wl.nlambda_h, output->wl.lambda_h, output->rfldn3d[lu][is][js], linear);

	      status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->flup3d_t[lu][is][js], 
				output->wl.nlambda_h, output->wl.lambda_h, output->flup3d[lu][is][js], linear);

	    }
	
	status += arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->atm.sza_t, 
			  output->wl.nlambda_h, output->wl.lambda_h, output->sza, linear);
	
	if (status!=0) {
	  fprintf (stderr, "Error %d returned by arb_wvn()\n", status);
	  return status;
	}
      }
      
      if (input.rte.solver == POLRADTRAN) {

	for (j=0;j<input.rte.nphi;j++) {
	  for (iu=0;iu<input.rte.nstr/2;iu++) {
	    for (is=0;is<input.rte.pol_nstokes;is++) {
	      
	      status = arb_wvn( output->wl.nlambda_t, output->wl.lambda_t, output->down_rad_t[lu][j][iu][is], 
				output->wl.nlambda_h, output->wl.lambda_h, output->down_rad[lu][j][iu][is], 
				linear);
	      if (status!=0) {
		fprintf (stderr, "Error %d returned by arb_wvn()\n", status);
		return status;
	      }
	      status = arb_wvn( output->wl.nlambda_t, output->wl.lambda_t, output->up_rad_t[lu][j][iu][is], 
				output->wl.nlambda_h, output->wl.lambda_h, output->up_rad[lu][j][iu][is], 
				linear);
	      if (status!=0) {
		fprintf (stderr, "Error %d returned by arb_wvn()\n", status);
		return status;
	      }
	    }
	  }
	}
      }
      else {
	for (iu=0;iu<input.rte.maxumu;iu++) {
	  status = arb_wvn( output->wl.nlambda_t, output->wl.lambda_t, output->u0u_t[lu][iu], 
			    output->wl.nlambda_h, output->wl.lambda_h, output->u0u[lu][iu], linear);
	  
	  if (status!=0) {
	    fprintf (stderr, "Error %d returned by arb_wvn()\n", status);
	    return status;
	  }
	}
	
	for (j=0;j<input.rte.nphi;j++) {
	  for (iu=0;iu<input.rte.maxumu;iu++) {
	    
	    status = arb_wvn( output->wl.nlambda_t, output->wl.lambda_t, output->uu_t[lu][j][iu], 
			      output->wl.nlambda_h, output->wl.lambda_h, output->uu[lu][j][iu], 
			      linear);
	    if (status!=0) {
	      fprintf (stderr, "Error %d returned by arb_wvn()\n", status);
	      return status;
	    }
	  }
	}
      }
    }

    if (input.rte.mc.passback3D && input.rte.mc.absorption)
      for (ks=0; ks<input.rte.mc.Nz_cloud; ks++)
	for (is=0; is<input.rte.mc.Nx_cloud; is++)
	  for (js=0; js<input.rte.mc.Ny_cloud; js++) {
	    status = arb_wvn(output->wl.nlambda_t, output->wl.lambda_t, output->abs3d_t[ks][is][js], 
			     output->wl.nlambda_h, output->wl.lambda_h, output->abs3d[ks][is][js], linear);
	    if (status!=0) {
	      fprintf (stderr, "Error %d returned by arb_wvn()\n", status);
	      return status;
	    }
	  }
  }

  return 0;
}





/**************************************************************/
/* Mutliply the transmittance with the extraterrestrial       */
/* irradiance.                                                */
/**************************************************************/

int multiply_extraterrestrial (input_struct input, output_struct *output)
{
  int lu=0, is=0, js=0, ks=0, iv=0, iu=0, j=0;

  /* multiply with extraterrestrial irradiance */

  if (input.transmittance == 0 && input.source!=SRC_THERMAL) {
    if ( input.rte.solver == POLRADTRAN ) {
      for (lu=0;lu<input.atm.nzout;lu++) {
	for (iv=0; iv<output->wl.nlambda_h; iv++) {
	  for (is=0;is<input.rte.pol_nstokes;is++) {
	    output->up_flux[lu][is][iv]  *= output->wl.fbeam[iv];
	    output->down_flux[lu][is][iv]  *= output->wl.fbeam[iv];
	    for (j=0;j<input.rte.nphi;j++) {
	      for (iu=0;iu<input.rte.nstr/2;iu++) {
		output->down_rad[lu][j][iu][is][iv] *= output->wl.fbeam[iv];
		output->up_rad[lu][j][iu][is][iv] *= output->wl.fbeam[iv];
	      }
	    }
	  }
	}
      }
    }
    else {
      
      for (iv=0; iv<output->wl.nlambda_h; iv++) {
	for (lu=0;lu<input.atm.nzout;lu++) {
	  
	  /* irradiance and actinic flux */
	  output->flup  [lu][iv]  *= output->wl.fbeam[iv];
	  output->rfldn [lu][iv]  *= output->wl.fbeam[iv];
	  output->rfldir[lu][iv]  *= output->wl.fbeam[iv];
	  output->uavg  [lu][iv]  *= output->wl.fbeam[iv];
	  output->uavgdn[lu][iv]  *= output->wl.fbeam[iv];
	  output->uavgso[lu][iv]  *= output->wl.fbeam[iv];
	  output->uavgup[lu][iv]  *= output->wl.fbeam[iv];
	  
	  /* 3D fields */
	  if (input.rte.mc.passback3D)
	    for (is=0; is<input.rte.mc.Nx_sample; is++)
	      for (js=0; js<input.rte.mc.Ny_sample; js++) {
		output->rfldir3d [lu][is][js][iv] *= output->wl.fbeam[iv];
		output->rfldn3d  [lu][is][js][iv] *= output->wl.fbeam[iv];
		output->flup3d   [lu][is][js][iv] *= output->wl.fbeam[iv];
	      }

	  for (iu=0;iu<input.rte.maxumu;iu++) {
	    
	    /* azimutally averaged radiance */
	    output->u0u[lu][iu][iv] *= output->wl.fbeam[iv];
	    
	    /* radiance */
	    for (j=0;j<input.rte.nphi;j++)
	      output->uu[lu][j][iu][iv] *= output->wl.fbeam[iv];
	  }
	}

	if (input.rte.mc.passback3D && input.rte.mc.absorption)
	  for (ks=0; ks<input.rte.mc.Nz_cloud; ks++) 
	    for (is=0; is<input.rte.mc.Nx_cloud; is++) 
	      for (js=0; js<input.rte.mc.Ny_cloud; js++) 
		output->abs3d [ks][is][js][iv] *= output->wl.fbeam[iv];
	  
      }
    }
  }
  return 0;
}

/***********************************************************************************/
/* Function: optical_properties                                           @61_30i@ */
/* Description:                                                                    */
/*  Calculates optical depth, single scattering albedo and                         */
/*  phase function from absorption and scattering cross sections                   */
/*  corresponding gas and particulate matter concentrations for                    */
/*  one wavelength.                                                                */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i61_30@ */
/***********************************************************************************/

int optical_properties (input_struct input, output_struct *output, int iv, int iq)
{
  static int first=1;
  int k=0, lu=0;
  int status=0;
  double bscar=0, babs=0, babsa=0, babsd=0, babsi=0, babso=0;
  double moma=0, momd=0, momi=0;
  double bext=0, bsca=0, bscaa=0, bscad=0, bscai=0, deltaz=0, ga=0, gd=0, gi=0;
  double sumsca=0;
  double p1=0, p2=0, p3=0, p4=0, p5=0, p6=0;
  FILE *f=NULL;

  int nlev = output->atm.nlev;
  int nlyr = output->atm.nlyr;

  double pressure=0, temperature=0;

  int new_rayleigh_method=1;  /* how to deal with Rayleigh in FU parameterization */
  float *tmp_rayleigh = calloc (nlyr, sizeof (float));



  if (first!=0) {
    first = 0;
    
    /* allocate memory for profiles */

    output->atm.cdento   = (float *) calloc (nlyr, sizeof(float));
    output->dtauc        = (float *) calloc (nlyr, sizeof(float));
    output->ssalb        = (float *) calloc (nlyr, sizeof(float));
    output->dtair        = (float *) calloc (nlev, sizeof(float));
    output->dtmol        = (float *) calloc (nlev, sizeof(float));
    output->dtaer        = (float *) calloc (nlev, sizeof(float));
    output->dtcld        = (float *) calloc (nlev, sizeof(float));
    output->dtice        = (float *) calloc (nlev, sizeof(float));
    output->omaer        = (float *) calloc (nlev, sizeof(float));
    output->omcld        = (float *) calloc (nlev, sizeof(float));
    output->omice        = (float *) calloc (nlev, sizeof(float));
    output->fgaer        = (float *) calloc (nlev, sizeof(float));
    output->fgcld        = (float *) calloc (nlev, sizeof(float));
    output->fgice        = (float *) calloc (nlev, sizeof(float));
    output->lwccld       = (float *) calloc (nlev, sizeof(float));
    output->recld        = (float *) calloc (nlev, sizeof(float));
    output->atm.zmc      = (float *) calloc (nlev, sizeof(float));
    output->atm.tempermc = (float *) calloc (nlev, sizeof(float));

    status = ASCII_calloc_float(&output->pmom, nlyr, input.rte.nstr+1);

    if (status!=0) {
      fprintf (stderr, " ... error %d allocating memory for output->pmom\n", status);
      return status;
    }


    /* Calculate column density of air (total density of the atmosphere). */
    /* Required to calculate Rayleigh optical depth.                      */

    if (input.rayleigh==RAYLEIGH_CALC) {
      for (lu=0; lu<nlyr; lu++) {
	
	/* The factor 1e5 converts z from km to cm.*/
	deltaz = (output->atm.zd[lu] - output->atm.zd[lu+1]) * 1e5;
      
	/* average air density over a layer assuming  dens = exp (-c * z) */
	if (output->atm.dens[MOL_AIR][lu] > 0 && output->atm.dens[MOL_AIR][lu+1] > 0 &&
	    output->atm.dens[MOL_AIR][lu] != output->atm.dens[MOL_AIR][lu+1]) {
	  if (fabs(1.0 - output->atm.dens[MOL_AIR][lu+1] / output->atm.dens[MOL_AIR][lu]) > EPSILON ) 
	    output->atm.cdento[lu] = 1.0 / (log(output->atm.dens[MOL_AIR][lu+1]/output->atm.dens[MOL_AIR][lu])) * 
	      (output->atm.dens[MOL_AIR][lu+1] - output->atm.dens[MOL_AIR][lu]) * deltaz ;
	  else
	    output->atm.cdento[lu] = 0.5*fabs((output->atm.dens[MOL_AIR][lu]+
					       output->atm.dens[MOL_AIR][lu+1]) * deltaz);
	}
	else
	  output->atm.cdento[lu] = 0.5*fabs((output->atm.dens[MOL_AIR][lu]+
					     output->atm.dens[MOL_AIR][lu+1]) * deltaz);
      }
    }
  }
  

  /* FU special: The Rayleigh optical depth depends on the subband */
  if (input.ck_scheme == CK_FU) {
#if HAVE_FULIOU
    status = ckdfuray (iv+1, iq+1, cos(output->atm.sza_t[iv]/180.0*PI), nlev,
		       tmp_rayleigh);
#else
    fprintf (stderr, "Error, Fu and Liou not supported!\n");
    return -1;
#endif

    if (status!=0) {
      fprintf (stderr, "error, status %d returned by ckdfuray()\n", status);
      return status;
    }
  }

  for (lu=0; lu<nlyr; lu++) {

    /* The factor 1.e+5 converts z from km to cm.*/
    deltaz= (output->atm.zd[lu] - output->atm.zd[lu+1])*1e5;

    bscar = 0;

    switch (input.rayleigh) {  /* if Rayleigh scattering is not switched off */
    case RAYLEIGH_CALC:
      if (input.ck_scheme != CK_FU)
	bscar = output->atm.cdento[lu] * output->crs.crs[0][MOL_AIR][iv];
      else {    /* Fu and Liou parameterization */
	
	if (new_rayleigh_method)
	  bscar = tmp_rayleigh[lu];
	else {
	  /* average air pressure over a layer assuming  press = exp (-c * z) */
	  if (output->atm.press[lu] > 0 && output->atm.press[lu+1] > 0) {
	    if (fabs(1.0 - output->atm.press[lu+1] / output->atm.press[lu]) > EPSILON )  {
	      pressure = 1.0 / (log(output->atm.press[lu+1]/output->atm.press[lu])) * 
		(output->atm.press[lu+1] - output->atm.press[lu]);
	    }
	    else
	      pressure = 0.5*fabs(output->atm.press[lu] + output->atm.press[lu+1]);
	  }
	  else 
	    pressure = 0.5*fabs(output->atm.press[lu] + output->atm.press[lu+1]);
	  
	  temperature = 0.5*fabs(output->atm.temper[lu] + output->atm.temper[lu+1]);
	  
	  bscar = fu_rayleigh (pressure, temperature, iv) * deltaz/100.0;
	}
      }
      break;

    case RAYLEIGH_FILE:
      bscar = output->atm.tau_rayleigh[lu];
      break;

    case RAYLEIGH_NONE:
      break;

    default:
      fprintf (stderr, "error, unknown Rayleigh option %d\n", input.rayleigh);
      return -1;
    }



    babso=0;

    switch (input.molabs) {
    case MOLABS_CALC:
      switch (input.ck_scheme) {
      case CK_KATO:
	
	/* CO2 absorption */
	babso += deltaz * output->crs.crs[lu][MOL_CO2][iv] * 
	  0.5 * (output->atm.dens[MOL_CO2][lu] + output->atm.dens[MOL_CO2][lu+1]);
	
	/* O3 absorption */
	babso += deltaz * output->crs.crs[lu][MOL_O3][iv]  *
	  0.5 * (output->atm.dens[MOL_O3][lu]  + output->atm.dens[MOL_O3][lu+1]);
	
	/* O2 absorption */
	babso += deltaz * output->crs.crs[lu][MOL_O2][iv]  *
	  0.5 * (output->atm.dens[MOL_O2][lu]  + output->atm.dens[MOL_O2][lu+1]);
	
	/* H2O absorption */
	babso += deltaz * output->crs.crs[lu][MOL_H2O][iv] *
	  0.5 * (output->atm.dens[MOL_H2O][lu] + output->atm.dens[MOL_H2O][lu+1]);
	
	break;

      case CK_FU:
      case CK_AVHRR_KRATZ:   /* ??? need to check if that is correct in the Kratz case ??? */
	/* kabs is in units of km-1 */
	babso += deltaz/1e5 * output->kabs.kabs[lu][iv];
	break;
	
      case CK_NONE:

	/* only ozone absorption */
	babso += 0.5*deltaz * (output->atm.dens[MOL_O3][lu]*output->crs.crs[lu][MOL_O3][iv]
			       + output->atm.dens[MOL_O3][lu+1]*output->crs.crs[lu+1][MOL_O3][iv]);
	break;
	
      default:
	fprintf (stderr, " ... error: unimplemented correlated-k scheme\n");
	return -1;
	break;
      }
      break;

    case MOLABS_FILE:
      babso = output->atm.tau_molabs[lu];
      break;

    case MOLABS_NONE:
      babso = 0;
      break;
      
    default:
      fprintf (stderr, "error, unknown molecular absorption option %d\n", input.molabs);
      return -1;
    }


    /* scattering by aerosols, clouds, and ice clouds */
    bscaa = output->aer.ssa[iv][lu] * output->aer.dtau[iv][lu];
    bscad = output->wc.ssa [iv][lu] * output->wc.dtau [iv][lu];
    bscai = output->ic.ssa [iv][lu] * output->ic.dtau [iv][lu];
	  
    ga    = output->aer.gg[iv][lu];
    gd    = output->wc.gg [iv][lu];
    gi    = output->ic.gg [iv][lu];

    /*  Phase function */
    sumsca =  bscar + bscad +  bscaa + bscai;
    if (sumsca < FLT_MIN)
      sumsca = FLT_MIN;


    /* moments of the phase functions */

    momd=bscad;
    moma=bscaa;
    momi=bscai;

    output->pmom[lu][0] = 1.0;

    for (k=1; k<=input.rte.nstr; k++) {

      /* aerosol */
      if (output->aer.nmom[iv] > 0) {
	if (k<output->aer.nmom[iv])
	  moma = bscaa * output->aer.moment[iv][k];
	else 
	  moma=0;
      }
      else 
	moma *= ga; /* aerosol,      HG */

      momd *= gd;   /* water clouds, HG */
      momi *= gi;   /* ice clouds,   HG */

      output->pmom[lu][k] = (momd + moma + momi) / sumsca;

      if (k == 2)   /* Rayleigh scattering */
	output->pmom[lu][k] += bscar/(sumsca*10.0);
    }

    /* absorption by aerosols, clouds, and ice clouds */
    babsa  = (1.0 - output->aer.ssa[iv][lu]) * output->aer.dtau[iv][lu];
    babsd  = (1.0 - output->wc.ssa [iv][lu]) * output->wc.dtau [iv][lu];
    babsi  = (1.0 - output->ic.ssa [iv][lu]) * output->ic.dtau [iv][lu];
    
    /* absorption */
    if (input.absorption) {
      babs = babsa + babso + babsd + babsi;
      if (babs > FLT_MAX) 
	babs = FLT_MAX;
    }
    else {
      babs  = 0;
      babso = 0;
      babsa = 0;
      babsd = 0;
      babsi = 0;
    }

    /* Scattering coefficient */
    bsca = bscar + bscad + bscaa + bscai;
    if (bsca > FLT_MAX) 
      bsca = FLT_MAX;
    
    /* Extinction coefficient */
    bext = babs + bsca;
    if (bext > FLT_MAX) 
      bext = FLT_MAX;


    /* Optical properties for the mc model */
    output->dtair[nlyr-1-lu] = bscar;
    output->dtmol[nlyr-1-lu] = babso;
    output->dtaer[nlyr-1-lu] = babsa + bscaa;
    output->dtcld[nlyr-1-lu] = babsd + bscad;
    output->dtice[nlyr-1-lu] = babsi + bscai;
		  	      
    output->omaer[nlyr-1-lu] = output->aer.ssa[iv][lu];
    output->omcld[nlyr-1-lu] = output->wc.ssa [iv][lu];
    output->omice[nlyr-1-lu] = output->ic.ssa [iv][lu];
    		  	      
    output->fgaer[nlyr-1-lu] = output->aer.gg[iv][lu];
    output->fgcld[nlyr-1-lu] = output->wc.gg [iv][lu];
    output->fgice[nlyr-1-lu] = output->ic.gg [iv][lu];

    /* effective droplet radius             */
    /* ??? this is a temporary solution ??? */
    if (input.wc.layer) {
      output->recld[nlyr-1-lu]  = output->wc.effr[lu+1];
      output->lwccld[nlyr-1-lu] = output->wc.lwc[lu+1];
    }
    else {
      output->recld[nlyr-1-lu] = 0.5 * (output->wc.effr[lu]+output->wc.effr[lu+1]);
      output->lwccld[nlyr-1-lu] = 0.5 * (output->wc.lwc[lu]+output->wc.lwc[lu+1]);
    }

    /* Set optical depth and single scattering albedo */
    if (bext < FLT_MIN) 
      bext = FLT_MIN;

    output->dtauc[lu] = bext;
    output->ssalb[lu] = bsca/bext;

    if ( 1.0-output->ssalb[lu] < FLT_MIN ) 
      output->ssalb[lu] = 1.0 - FLT_MIN;
    
    if (input.rte.solver == POLRADTRAN ) {
      if ( (f = fopen(&output->atm.pol_scat_files[lu*64], "w")) == NULL) return 1;
      fprintf(f,"%e\n",bext);
      fprintf(f,"%e\n",bsca);
      fprintf(f,"%e\n",bsca/bext);

      fprintf(f,"%d\n", input.rte.nstr); 
      momd=bscad;
      moma=bscaa;
      momi=bscai;
      for (k=0; k<=input.rte.nstr; k++) { 
	if (output->aer.nmom[iv] > 0) {
	  if (k<output->aer.nmom[iv])
	    moma = bscaa * output->aer.moment[iv][k];
	  else 
	    moma=0;
	}
	else 
	  moma *= ga;   /* aerosol,      HG */

	momd *= gd;   /* water clouds, HG */
	momi *= gi;   /* ice clouds,   HG */
	/* The first numbers in the columns are the Legendre coefficients
	   for the Rayleigh scattering phase matrix. Note that disort wants
	   phase function moments g_l while polradtran wants Legendre 
	   coefficients C_l = (2*l+1)g_l. p1 is the scalar story of everything. */
	if (k==0) {     
	  p1 =  1.0;
	  p2 = -0.5;
	  p3 =  0.0;
	  p4 =  0.0;
	  p5 =  1.0;
	  p6 =  0.0;
	} 
	else if (k==1) { 
	  p1 =  (0.0 + (2*k+1)*(momd + moma + momi) )/ sumsca; 
	  p2 =  0.0; 
	  p3 =  1.5; 
	  p4 =  0.0; 
	  p5 =  0.0; 
	  p6 =  1.5; 
	} 
	else if (k==2) { 
	  p1 =  (0.5*bscar + (2*k+1)*(momd + moma + momi) )/ sumsca; 
	  p2 =  0.5; 
	  p3 =  0.0; 
	  p4 =  0.0; 
	  p5 =  0.5; 
	  p6 =  0.0; 
	}
	else {
	  p1 =  (0.0+ (2*k+1)*(momd + moma + momi) )/ sumsca; 
	  p2 =  0.0; 
	  p3 =  0.0; 
	  p4 =  0.0; 
	  p5 =  0.0; 
	  p6 =  0.0; 
	} 
	fprintf(f,"%d %f %f %f %f %f %f\n", k, p1, p2, p3, p4, p5, p6); 
      }
      fclose(f);
    }
  }


  output->dtair[nlyr] = 0.0;
  output->dtmol[nlyr] = 0.0;
  output->dtaer[nlyr] = 0.0; 
  output->dtcld[nlyr] = 0.0;
  output->dtice[nlyr] = 0.0;
                     
  output->omaer[nlyr] = 0.0;
  output->omcld[nlyr] = 0.0;
  output->omice[nlyr] = 0.0;
                     
  output->omaer[nlyr] = 0.0;
  output->omcld[nlyr] = 0.0;
  output->omice[nlyr] = 0.0;

  for (lu=0; lu<=nlyr; lu++) {
    output->atm.zmc[nlyr-lu] = output->atm.zd[lu];
    output->atm.tempermc[nlyr-lu] = output->atm.temper[lu];
  }

  
	   

  if (input.ck_scheme == CK_FU)
    free (tmp_rayleigh);

  return 0;
}



/***********************************************************************************/
/* Function: crs_rayleigh                                                          */
/* Description:                                                                    */
/*    Calculates the Rayleigh scattering cross section according to the formula    */
/*    by Nicolet, M., `On the molecular scattering in the terrestrial atmosphere:  */
/*    an empirical formula for its calculation in the homosphere', Planet. Space   */
/*    Sci., 32, 1467-1468, 1984.                                                   */
/*                                                                                 */
/* Parameters:                                                                     */
/*  float* lambda:         Pointer to ary holding the wavelength in nanometers.    */
/*  int n_lambda:          Number of wavelength points.                            */
/*                                                                                 */
/* Return value:                                                                   */
/*  double pointer to ary holding the Rayleigh scattering cross section as a       */
/*  function of wavelength.                                                        */
/*                                                                                 */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                                 */
/***********************************************************************************/

int crs_rayleigh (float *lambda, int n_lambda, float ***crs) 
{
  float ray_const  = 4.02e-28;
  float from_nm_um = 1.0e-3;
  float lambda_um=0.0, x=0.0;
  int iv=0;
  
  for (iv=0; iv<n_lambda; iv++) {

    lambda_um = lambda[iv] * from_nm_um;

    if (lambda_um > 0.55)
      x = 0.04;
    else
      x = 0.389*lambda_um + 0.09426/lambda_um - 0.3228;
    
    crs[0][MOL_AIR][iv] = ray_const / pow(lambda_um,4+x);
  }

  return 0;
}



/***********************************************************************************/
/* Sum the 3D results over wavelength and print data to mc.sum                     */ 
/***********************************************************************************/

int integrate3D (input_struct input, output_struct *output)
{
  int is=0, js=0, ks=0, iv=0;
  double rfldir=0, rfldn=0, flup=0, absorption=0;

  FILE *f    = fopen ("mc.sum", "w");
  FILE *fabs = fopen ("mcabs.sum", "w");


  for (ks=0; ks<input.atm.nzout; ks++) {
    for (is=0; is<input.rte.mc.Nx_sample; is++) {
      for (js=0; js<input.rte.mc.Ny_sample; js++) {
	
	rfldir=0;
	rfldn=0;
	flup=0;
	
	for (iv=0;iv<output->wl.nlambda_h;iv++) {
	  rfldir += output->rfldir3d[ks][is][js][iv];
	  rfldn  += output->rfldn3d [ks][is][js][iv];
	  flup   += output->flup3d  [ks][is][js][iv];
	}
	
	fprintf (f, "%d %d %d %g %g %g\n", is, js, ks, rfldir, rfldn, flup);
      }
    }
  }
  

  if (input.rte.mc.passback3D && input.rte.mc.absorption)
    for (ks=0; ks<input.rte.mc.Nz_cloud; ks++)
      for (is=0; is<input.rte.mc.Nx_cloud; is++)
	for (js=0; js<input.rte.mc.Ny_cloud; js++) {
	  absorption = 0;
	  for (iv=0;iv<output->wl.nlambda_h;iv++)
	    absorption += output->abs3d[ks][is][js][iv];
	  
	  fprintf (fabs, "%d %d %d %g\n", is, js, ks, absorption);
	}

  
  
  fclose(f); fclose(fabs);

  return 0;
}
	  


