/************************************************************************/
/*                                                                      */
/* Copyright (C) 1997 Arve Kylling                                      */
/*                                                                      */
/* This program is free software; you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation; either version 1, or (at your option)  */
/* any later version.                                                   */
/*                                                                      */
/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY of FITNESS FOR A PARTICULAR PURPOSE. See the         */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* To obtain a copy of the GNU General Public License write to the      */
/* Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,   */
/* USA.                                                                 */
/*                                                                      */
/************************************************************************/

#ifndef __uvspec_h
#define __uvspec_h

#include <stdio.h>

#define VERSION 0.99

/* Indices for the reading of the atmosphere files */
#define ATMFILE_ZD       0
#define ATMFILE_PRESS    1
#define ATMFILE_TEMPER   2
#define ATMFILE_AIR      3
#define ATMFILE_O3       4
#define ATMFILE_O2       5
#define ATMFILE_H2O      6
#define ATMFILE_CO2      7
#define ATMFILE_NO2      8


#define MOL_NN        6
#define MOL_AIR       0
#define MOL_O3        1
#define MOL_O2        2
#define MOL_H2O       3
#define MOL_CO2       4
#define MOL_NO2       5


#define FN_NN               13
#define FN_PATH              0
#define FN_ALBEDO            1
#define FN_ATMOSPHERE        2
#define FN_EXTRATERRESTRIAL  3
#define FN_SZA               4
#define FN_WLTRANS           5
#define FN_WCLOUD            6
#define FN_ICLOUD            7
#define FN_SLITFUNCTION      8
#define FN_SPLINE            9
#define FN_RAYLEIGH         10
#define FN_MOLABS           11
#define FN_RPV              12

#define FN_AER_NN            6
#define FN_AER_TAU           0
#define FN_AER_GG            1
#define FN_AER_SSA           2
#define FN_AER_MOM           3
#define FN_AER_REF           4
#define FN_AER_SIZ           5

#define FN_MC_NN             4
#define FN_MC_PHOTONS        0
#define FN_MC_MOMCLD         1
#define FN_MC_MOMTRNC        2
#define FN_MC_PCLD           3

#define CK_NONE         0
#define CK_KATO         1
#define CK_FU           2
#define CK_AVHRR_KRATZ  3

#define SRC_NONE    0
#define SRC_SOLAR   1
#define SRC_THERMAL 2

#define O3_MOLINA    1
#define O3_DAUMONT   2
#define O3_BASSPAUR  3    

#define RAYLEIGH_NONE 0
#define RAYLEIGH_CALC 1
#define RAYLEIGH_FILE 2

#define MOLABS_NONE 0
#define MOLABS_CALC 1
#define MOLABS_FILE 2

#define MX_NN  7  
#define MX_CO2 0
#define MX_H2O 1
#define MX_CH4 2
#define MX_N2O 3
#define MX_F11 4
#define MX_F12 5
#define MX_F22 6

#define WCZD     0    
#define WCLWC    1
#define WCEFFR   2

#define ICZD     0    
#define ICLWC    1
#define ICEFFR   2

#define DISORT     1
#define SDISORT    2
#define TWOSTR     3
#define SUCORD     4
#define MONTECARLO 5
#define POLRADTRAN 6

#define MOMCLD_NONE     0
#define MOMCLD_DEFAULT  1
#define MOMCLD_WATER    2
#define MOMCLD_ICE      3
#define MOMCLD_FILE     4

#define MAXAWVN 47


/* minimum number of photons for each band*/
#define MIN_MCPHOTONS 100


/* Monte Carlo input STRUCTURE */

typedef struct {
  int     photons;

  int     Nx_sample;
  int     Ny_sample;

  int     Nx_cloud;
  int     Ny_cloud;
  int     Nz_cloud;

  int     passback3D;
  int     ipa;         /* MYSTIC in IPA mode                 */
  int     escape;      /* escape probabilities for radiances */
  int     absorption;  /* calculate absorption for each box  */ 

  int     momcld;
  char    **filename;
} mc_inp_struct;


/* Raditive Transfer Equation STRUCTURE */
/* See disort.doc for description of variables */

typedef struct {
  int     deltam;
  int     fisot;
  int     nstr;  
  int     nscat;
  int     maxphi;
  int     maxumu;
  int     nphi;
  int     numu;
  int     solver;
  int     nprndis;
  int     reverse;
  int*    prndis;
  int     pol_aziorder;
  int     pol_nstokes;
  char*   pol_quad_type;
  char    pol_stokes[4];
  int     pol_src_code;
  double  pol_max_delta_tau;
  float*  phi;
  float*  umu;
  mc_inp_struct mc;
} rte_struct;

/* Aerosol input STRUCTURE */

typedef struct {
  char **filename;   /* array of filenames */
  int haze;
  int vulcan;
  int seasn;
  float alpha;
  float beta;
  float visib;
  
  float re;          /* refractive index            */
  float im;          /* (independent of wavelength) */

  float scale_ssa;
  float set_gg;
  float set_ssa;
} aer_inp_struct;

/* Aerosol output STRUCTURE */

typedef struct {
  float   *aer_pro;
  float   **dtau;
  float   **gg;
  float   **ssa;

  double   *xsiz;    /* size distribution              */
  double   *ysiz;
  int      nsiz;

  float   *re_t;     /* real and imaginary part of the */
  float   *im_t;     /* refractive index               */

  float   **moment;  /* moments of the phase function for each wavelength */
  int     *nmom;     /* number of moments for each wavelength             */
} aer_out_struct;



/* RPV (Rahman, Pinty, Verstraete) structure */

typedef struct {
  float rho0;
  float k;
  float theta;
} rpv_inp_struct;


/* Cloud input STRUCTURE */

typedef struct {
  int     layer;
  float   tau;  
  float   gg;  
  float   ssa;  
} cld_inp_struct;



/* RPV (Rahman, Pinty, Verstraete) structure */
/* to store wavelength dependent BRDF        */
typedef struct {
  int   n;
  float *lambda;
  float *rho0;     /* input wavelength grid    */
  float *k;
  float *theta;

  float *rho0_t;   /* internal wavelength grid */
  float *k_t;
  float *theta_t;
} rpv_out_struct;



/* Cloud output STRUCTURE */

typedef struct {
  float   **dtau;
  float   **gg;
  float   **ssa;
  float   *lwc;
  float   *effr;
  float   *zd;
  int     nlev;
  int     nlam;
  int     newsiz;
} cld_out_struct;

/* Wavelength input grid STRUCTURE */

typedef struct {
  float   wl_start;
  float   wl_end;
} wl_inp_struct;

/* Wavelength output grid STRUCTURE */

typedef struct {
  int   nlambda_h;
  int   nlambda_s;
  int   nlambda_t;

  int   *map_e2h;

  float *lambda_h;
  float *lambda_s;
  float *lambda_t;
  float *fbeam;
  float *wvnmlo_t;
  float *wvnmhi_t;
} wl_out_struct;

/* Atmosphere input data STRUCTURE */

typedef struct {
  float *zout;
  int    nzout;
  float  albedo;
  float  altitude;
  float  sza;
  float  phi0;
} atm_inp_struct;

/* Atmosphere output data STRUCTURE */

typedef struct {
  char* pol_scat_files;
  float *cdento;
  float **dens;
  float *press;
  float *rh;
  float *temper;
  float *temper_out;
  float *zd;
  float *zmc;
  float *tempermc;
  float *tau_rayleigh;
  float *tau_molabs;
  float *file_albedo;
  float *file_lambda_albedo;
  float *albedo_t;
  float *file_phi0;
  float *file_sza;
  float *file_lambda_sza;
  float *sza_t;
  float *phi0_t;
  int    file_nlambda_albedo;
  int    nlambda_albedo;
  int    file_nlambda_sza;
  int    nlambda_sza;
  int    nlev;
  int    nlyr;
  int    ndens;
} atm_out_struct;

/* Cross section input data STRUCTURE */

typedef struct {
  char*  filename;
  int    o3_type;
  int    column;
} crs_inp_struct;


/*******************************************************************/
/* The ck structure holds all the information required for         */
/* a correlated-k approximation.                                   */
/* Components:                                                     */
/*    scheme               CK_KATO, CK_FU                          */
/*    wvlc [1 ... n_wvl]   central wavelength [nm]                 */                   
/*    wvl1 [1 ... n_wvl]   start wavelength [nm] for each band     */    
/*    wvl2 [1 ... n_wvl]   end wavelength [nm] for each band       */    
/*    n_wvl                number of wavelength bands              */
/*    comp [1 ... n_comp][1 ... n_wvl]                             */
/*                         monochromatic cross sections bands      */
/*                         which don't require c-k but only a      */
/*                         monochromatic calculation               */
/*                         (e.g. ozone in the UV/VIS)              */
/*    wght [1 ... n_comp][1 ... n_wvl]                             */
/*                         number of quadrature points             */
/*******************************************************************/

typedef struct {   /* correlated-k structure */
  int scheme;

  double *wvlc;
  double *wvl1;
  double *wvl2;
  int n_wvl;

  double **comp;
  int    **wght;

  void *x_thermal;  /* photon fraction; x_thermal[iv][iq1][iq2][...] */
  void *x_solar;    /* photon fraction; x_solar  [iv][iq1][iq2][...] */

  int n_comp;
} ck_struct;

/* Cross section output data STRUCTURE */

typedef struct {
  float*** crs; /* 3D ary, for cross section[temperature (or layer), type, wavelength]*/
} crs_out_struct;

typedef struct {
  float** kabs; /* 2D ary, for absorption coefficient[temperature (or layer), wavelength] */
} kabs_out_struct;

typedef struct {
  double **crs;      /* cross section [temperature, quadrature point] */
  double *weight;    /* weight [quadrature point]                     */
  int ngauss;        /* number of quadrature points                   */
} ck_profile;

typedef struct {
  ck_profile **profile;  /* 2D ary, for ck_profile [type, band] */
} crs_ck_out_struct;


			    
/* INPUT STRUCTURE */

typedef struct {
  char         **filename;   /* array of filenames */
  
  float        cloudcover;
  int          ipa;
  int          convolve;
  int          spline;
  int          header;
  int          flexstor;
  int          source;
  int          rayleigh;
  int          molabs;
  int          absorption;
  float        spline_lambda_0;
  float        spline_lambda_1;
  float        spline_lambda_step;
  int          day_of_year;
  float        ozone_column;
  float        pressure;
  float        *mixing_ratio;
  int transmittance;
  rte_struct   rte;
  atm_inp_struct atm;
  aer_inp_struct aer;
  crs_inp_struct crs;
  rpv_inp_struct rpv;
  wl_inp_struct  wl;
  cld_inp_struct wc;
  cld_inp_struct ic;
  int            ck_scheme;
  int            ck_h2ocont;
} input_struct;


/* OUTPUT STRUCTURE */

typedef struct {
  int         phi;
} print_struct;

/* elements whose memory requirements have been free'ed are listed behind */
/* each declaration. Once they are all done rm this comment.              */

typedef struct {
  float       eccentricity_factor;
  float       ozone_column;
  float       ozone_scale_factor;

  int         mc_photons;
  char        *mc_momcldfilename;

  float       *mixing_ratio;

  float       *dtair;           /* free'ed */
  float       *dtmol;           /* free'ed */
  float       *dtaer;           /* free'ed */
  float       *dtcld;           /* free'ed */
  float       *dtice;           /* free'ed */
  float       *omaer;           /* free'ed */
  float       *omcld;           /* free'ed */
  float       *omice;           /* free'ed */
  float       *fgaer;
  float       *fgcld;
  float       *fgice;

  float       *recld;
  float       *lwccld;

  float       *dtauc;           /* free'ed */
  float       *utau;
  float       *ssalb;
  float       **pmom;

  /* flux, actinic flux, and intensities on the output grid */ 
  float       **rfldir;     /* direct flux                    */
  float       **rfldn;      /* diffuse downward flux          */
  float       **flup;       /* diffuse upward flux            */ 

  float       **uavgso;     /* direct actinic flux            */
  float       **uavgdn;	    /* diffuse downward actinic flux  */
  float       **uavgup;	    /* diffuse upward actinic flux    */

  float       **uavg;       /* average intensity              */
  float       ***u0u;       /* azimutally averaged intensites */
  float       ****uu;       /* intensities                    */

  /* MYSTIC 3D fields on the output grid */  
  float       ****rfldir3d; 
  float       ****rfldn3d;
  float       ****flup3d;
  float       ****abs3d;

  /* flux, actinic flux, and intensities on the internal transmittance grid */ 
  float       **rfldir_t;  /* direct flux                    */
  float       **rfldn_t;   /* diffuse downward flux          */
  float       **flup_t;	   /* diffuse upward flux            */
			                                       
  float       **uavgso_t;  /* direct actinic flux            */
  float       **uavgdn_t;  /* diffuse downward actinic flux  */
  float       **uavgup_t;  /* diffuse upward actinic flux    */
			                                       
  float       **uavg_t;	   /* average intensity              */
  float       ***u0u_t;	   /* azimutally averaged intensites */
  float       ****uu_t;	   /* intensities                    */

  /* MYSTIC 3D fields on the internal transmittance grid */  
  float       ****rfldir3d_t;
  float       ****rfldn3d_t;
  float       ****flup3d_t;
  float       ****abs3d_t;


  /* PolRadtran flux and intensities on the output grid */ 
  float       ***down_flux;     /* free'ed */
  float       ***up_flux;       /* free'ed */
  float       *****down_rad;    /* free'ed */
  float       *****up_rad;      /* free'ed */

  /* PolRadtran flux and intensities on the internal transmittance grid */ 
  float       ***down_flux_t;   /* free'ed */
  float       ***up_flux_t;     /* free'ed */
  float       *****down_rad_t;  /* free'ed */
  float       *****up_rad_t;    /* free'ed */


  float       *mu_values;       /* free'ed */
  float       *sza;             /* free'ed */

  atm_out_struct    atm;
  aer_out_struct    aer;
  crs_out_struct    crs;
  rpv_out_struct    rpv;
  kabs_out_struct   kabs;
  crs_ck_out_struct crs_ck;

  cld_out_struct    wc;
  cld_out_struct    *wcipa;
  cld_out_struct    ic;
  float             *ipaweight;
  int               nipa;

  wl_out_struct     wl;
  print_struct      print;
  ck_struct         ck;
} output_struct;

int uvspec(input_struct input, output_struct *output);
int uvspec_check(input_struct input);


/* include all uvspec header files */
#include "atmosphere.h"
#include "extraterrestrial.h"
#include "sza.h"
#include "albedo.h"
#include "aerosol.h"
#include "cloud.h"
#include "molecular.h"
#include "ancillary.h"
#include "solve_rte.h"
#include "ipa.h"
#include "ck.h"
#endif



